# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the @dataobject decorator"""

# Standard
from dataclasses import dataclass, field, is_dataclass
from enum import Enum
from typing import Dict, List, Optional, Union
import copy
import json
import os
import tempfile

# Third Party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message
import numpy as np
import pytest

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber, OneofField

# Local
from caikit.core import (  # NOTE: Imported from the top to validate
    DataObjectBase,
    dataobject,
)
from caikit.core.data_model import enums
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.data_backends.dict_backend import DictBackend
from caikit.core.data_model.dataobject import (
    CAIKIT_DATA_MODEL,
    make_dataobject,
    render_dataobject_protos,
)
from caikit.core.data_model.enums import isprotobufenum
from caikit.core.data_model.json_dict import JsonDict
from tests.data_model_helpers import reset_global_protobuf_registry, temp_dpool

## Helpers #####################################################################


@pytest.fixture(autouse=True)
def auto_temp_dpool():
    """Fixture to isolate the descriptor pool used in each test"""
    with temp_dpool() as dpool:
        yield dpool


@pytest.fixture(autouse=True)
def auto_reset_global_protobuf_registry():
    """Reset the global registry of generated protos"""
    with reset_global_protobuf_registry():
        yield


def check_field_type(proto_class, field_name, exp_type):
    field = proto_class.DESCRIPTOR.fields_by_name[field_name]
    return field.type == getattr(field, exp_type)


def check_field_label(proto_class, field_name, exp_label):
    field = proto_class.DESCRIPTOR.fields_by_name[field_name]
    return field.label == getattr(field, exp_label)


def check_field_message_type(proto_class, field_name, exp_msg_descriptor):
    field = proto_class.DESCRIPTOR.fields_by_name[field_name]
    return field.message_type is exp_msg_descriptor


def check_field_enum_type(proto_class, field_name, exp_enum_descriptor):
    field = proto_class.DESCRIPTOR.fields_by_name[field_name]
    return field.enum_type is exp_enum_descriptor


## Tests #######################################################################


def test_dataobject_bytes():
    @dataobject
    class Foo(DataObjectBase):
        foo: bytes

    f1 = Foo(b"asdf")
    assert f1.foo == b"asdf"
    assert f1.to_json() == '{"foo": "YXNkZg=="}'
    # test round-trip json
    f2 = Foo.from_json(f1.to_json())
    assert f2.foo == f1.foo
    assert f2.to_json() == f1.to_json()


def test_dataobject_bytes_union():
    @dataobject
    class Foo(DataObjectBase):
        foo: Union[bytes, int]

    f1 = Foo(b"asdf")
    assert f1.foo == b"asdf"
    assert f1.to_json() == '{"foo_bytes": "YXNkZg=="}'
    f2 = Foo(1)
    assert f2.foo == 1
    assert f2.to_json() == '{"foo_int": 1}'
    # # test round-trip json
    f3 = Foo.from_json(f1.to_json())
    assert f3.foo == f1.foo
    assert f3.to_json() == f1.to_json()


def test_dataobject_native_types():
    """Make sure that a simple usage of dataobject for a flat object works with
    fields declared using  native python types works
    """

    @dataobject
    class Foo(DataObjectBase):
        foo: str
        bar: int

    assert check_field_type(Foo.get_proto_class(), "foo", "TYPE_STRING")
    assert check_field_type(Foo.get_proto_class(), "bar", "TYPE_INT64")

    inst = Foo(foo="test", bar=1)
    assert inst.foo == "test"
    assert inst.bar == 1
    inst = Foo()
    assert inst.foo is None
    assert inst.bar is None


def test_dataobject_nested_objects():
    """Make sure that nested objects are handled correctly"""

    @dataobject
    class Foo(DataObjectBase):
        @dataobject
        class Bar(DataObjectBase):
            bat: str

        bar: Bar

    assert hasattr(Foo, "Bar")
    bar_type = Foo.Bar
    assert issubclass(bar_type, DataBase)
    assert issubclass(bar_type.get_proto_class(), message.Message)
    assert check_field_type(Foo.get_proto_class(), "bar", "TYPE_MESSAGE")


def test_dataobject_nested_enums():
    """Make sure enums work as nested fields"""

    @dataobject
    class Foo(DataObjectBase):
        @dataobject
        class Bar(Enum):
            EXAM = 0
            DRINKS = 1

        @dataobject
        class Bat(DataObjectBase):
            @dataobject
            class Kind(Enum):
                BASEBALL = 0
                VAMPIRE = 1

            kind: Kind

        bar: Bar
        bat: Bat

    assert hasattr(Foo, "Bat")

    # Foo.Bar is a nested enum
    bar_type = Foo.Bar
    assert issubclass(bar_type, Enum)
    assert getattr(enums, "Bar") is bar_type

    # Foo.Bat is a nested message
    bat_type = Foo.Bat
    assert issubclass(bat_type, DataBase)
    assert issubclass(bat_type.get_proto_class(), message.Message)
    assert check_field_type(Foo.get_proto_class(), "bat", "TYPE_MESSAGE")

    # Foo.Bat.Kind is a nested-nested enum
    kind_type = Foo.Bat.Kind
    assert issubclass(kind_type, Enum)
    assert getattr(enums, "Kind") is kind_type


def test_dataobject_top_level_enums():
    """Make sure enums can be created"""

    @dataobject
    class Foo(Enum):
        EXAM = 0
        DRINKS = 1

    # Make sure it is in fact a class! This is surprisingly hard to achieve.
    assert isinstance(Foo, type)
    assert Foo.EXAM.value == 0
    assert Foo.DRINKS.value == 1
    dict_view = Foo.to_dict()
    assert dict_view == {"EXAM": 0, "DRINKS": 1}

    # Make sure the underlying proto enum is accessible
    proto_enum = Foo._proto_enum
    assert isprotobufenum(proto_enum)


def test_dataobject_arrays():
    """Make sure arrays work as expected"""

    @dataobject
    class Foo(DataObjectBase):
        bar: List[str]

    assert check_field_type(Foo.get_proto_class(), "bar", "TYPE_STRING")
    assert check_field_label(Foo.get_proto_class(), "bar", "LABEL_REPEATED")


def test_dataobject_simple_dict():
    """Make sure simple dicts work as expected"""

    @dataobject
    class Foo(DataObjectBase):
        bar: Dict[str, int]

    assert check_field_type(Foo.get_proto_class(), "bar", "TYPE_MESSAGE")
    assert (
        Foo.get_proto_class()
        .DESCRIPTOR.fields_by_name["bar"]
        .message_type.GetOptions()
        .map_entry
    )

    dict_input = {"foo": 1, "bar": 2}
    foo = Foo(bar=dict_input)
    assert foo.bar == dict_input

    assert Foo.from_proto(foo.to_proto()).bar == dict_input


def test_dataobject_obj_refs_no_opt_types():
    """Make sure that references to other data objects and enums work as
    expected
    """

    @dataobject
    class BarEnum(Enum):
        EXAM = 0
        METAL = 1

    @dataobject
    class Foo(DataObjectBase):
        foo: str

    @dataobject
    class FooBar(DataObjectBase):
        foo: Foo
        bar: BarEnum

    assert check_field_type(FooBar.get_proto_class(), "foo", "TYPE_MESSAGE")
    assert check_field_message_type(
        FooBar.get_proto_class(), "foo", Foo.get_proto_class().DESCRIPTOR
    )
    assert check_field_enum_type(
        FooBar.get_proto_class(), "bar", BarEnum._proto_enum.DESCRIPTOR
    )


def test_dataobject_obj_refs_with_optional_types():
    """Make sure that references to other data objects and enums work as
    expected
    """

    @dataobject
    class BarEnum(Enum):
        EXAM = 0
        METAL = 1

    @dataobject
    class Foo(DataObjectBase):
        foo: str

    @dataobject
    class FooBar(DataObjectBase):
        foo: Foo
        optionalFoo: Optional[Foo]
        bar: BarEnum
        optionalBar: Optional[BarEnum]

    assert check_field_type(FooBar._proto_class, "foo", "TYPE_MESSAGE")
    for field in ["foo", "optionalFoo"]:
        assert check_field_message_type(
            FooBar._proto_class, field, Foo._proto_class.DESCRIPTOR
        )
    for field in ["bar", "optionalBar"]:
        assert check_field_enum_type(
            FooBar._proto_class, field, BarEnum._proto_enum.DESCRIPTOR
        )


def test_dataobject_additional_methods():
    """Make sure that additional methods on wrapped classes (for messages and
    enums) are preserved
    """

    @dataobject
    class Foo(Enum):
        EXAM = 0
        DRINKS = 1

        @classmethod
        def is_exam(cls, val: "Foo") -> bool:
            return val == cls.EXAM

    @dataobject
    class Bar(DataObjectBase):
        bar: str

        def caps(self) -> str:
            return self.bar.upper()

    assert Foo.is_exam(Foo.EXAM)
    assert Bar("bat").caps() == "BAT"


def test_render_dataobject_protos_valid_dir():
    """Make sure that render_dataobject_protos correctly renders all generated
    protobufs to the target directory
    """

    @dataobject
    class BarEnum(Enum):
        EXAM = 0
        METAL = 1

    @dataobject
    class Foo(DataObjectBase):
        foo: str

    @dataobject
    class FooBar(DataObjectBase):
        foo: Foo
        bar: BarEnum

    with tempfile.TemporaryDirectory() as workdir:
        render_dataobject_protos(workdir)
        rendered_files = set(os.listdir(workdir))
        assert rendered_files == {
            BarEnum._proto_enum.DESCRIPTOR.file.name,
            Foo.get_proto_class().DESCRIPTOR.file.name,
            FooBar.get_proto_class().DESCRIPTOR.file.name,
        }


def test_render_dataobject_protos_no_dir():
    """Make sure that render_dataobject_protos correctly renders all generated
    protobufs to the target directory and creates the target dir if it doesn't exist
    """

    @dataobject
    class BarEnum(Enum):
        EXAM = 0
        METAL = 1

    @dataobject
    class Foo(DataObjectBase):
        foo: str

    @dataobject
    class FooBar(DataObjectBase):
        foo: Foo
        bar: BarEnum

    with tempfile.TemporaryDirectory() as workdir:
        protos_dir_path = os.path.join(workdir, "protos")
        render_dataobject_protos(protos_dir_path)
        rendered_files = set(os.listdir(protos_dir_path))
        assert rendered_files == {
            BarEnum._proto_enum.DESCRIPTOR.file.name,
            Foo.get_proto_class().DESCRIPTOR.file.name,
            FooBar.get_proto_class().DESCRIPTOR.file.name,
        }


def test_dataobject_with_oneof():
    """Make sure that using a Union to create a oneof works as expected"""

    @dataobject
    class BazObj(DataObjectBase):
        @dataobject
        class Foo(DataObjectBase):
            data: List[str]

        @dataobject
        class Bar(DataObjectBase):
            data: str

        data_stream: Union[
            Annotated[Foo, FieldNumber(1), OneofField("foo")],
            Annotated[Bar, FieldNumber(2), OneofField("bar")],
        ]

    # Construct with oneof field name
    foo1 = BazObj(foo=BazObj.Foo(data=["hello"]))
    assert isinstance(foo1.data_stream, BazObj.Foo)
    assert foo1.which_oneof("data_stream") == "foo"
    assert foo1.foo is foo1.data_stream
    assert foo1.bar is None

    # Test other oneof field name
    bar1 = BazObj(bar=BazObj.Bar(data="world"))
    assert isinstance(bar1.data_stream, BazObj.Bar)
    assert bar1.which_oneof("data_stream") == "bar"
    assert bar1.bar is bar1.data_stream
    assert bar1.foo is None

    # Test to_dict
    dict_repr_foo = foo1.to_dict()
    assert dict_repr_foo == {
        "foo": {"data": ["hello"]},
    }
    assert dict_repr_foo["foo"]["data"] == ["hello"]

    # Test proto round trip
    proto_repr_foo = foo1.to_proto()
    assert proto_repr_foo.foo.data == ["hello"]
    assert BazObj.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo
    proto_repr_bar = bar1.to_proto()
    assert proto_repr_bar.bar.data == "world"

    # Test json round trip
    json_repr_foo = foo1.to_json()
    assert json.loads(json_repr_foo) == {
        "foo": {"data": ["hello"]},
    }
    assert BazObj.from_json(json_repr_foo) == foo1

    # Test setattr
    foo1.bar = BazObj.Bar(data="it's a bar")
    assert foo1.which_oneof("data_stream") == "bar"
    assert foo1.data_stream is foo1.bar
    assert foo1.foo is None

    # Construct with oneof name
    foo2 = BazObj(data_stream=BazObj.Foo(data=["some", "foo"]))
    assert foo2.data_stream.data == ["some", "foo"]
    assert foo2.bar is None
    assert foo2.foo is foo2.data_stream
    assert foo2.which_oneof("data_stream") == "foo"

    # Assign with oneof name
    foo2.data_stream = BazObj.Bar(data="asdf")
    assert foo2.foo is None
    assert foo2.bar is foo2.data_stream
    assert foo2.which_oneof("data_stream") == "bar"

    # Construct with positional oneof name
    foo2 = BazObj(BazObj.Foo(data=["some", "foo"]))
    assert foo2.data_stream.data == ["some", "foo"]
    assert foo2.bar is None
    assert foo2.foo is foo2.data_stream
    assert foo2.which_oneof("data_stream") == "foo"

    foo3 = BazObj()
    assert foo3.foo is None
    assert foo3.bar is None
    assert foo3.data_stream is None
    assert foo3.which_oneof("data_stream") == None
    # Invalid constructors
    with pytest.raises(TypeError):
        BazObj(BazObj.Foo(), foo=BazObj.Foo())
    with pytest.raises(TypeError):
        BazObj(data_stream=BazObj.Foo(), foo=BazObj.Foo())
    with pytest.raises(TypeError):
        BazObj(foo=BazObj.Foo(), bar=BazObj.Bar())


def test_dataobject_with_same_type_of_oneof():
    """Make sure that using a Union to create a oneof with the same types works as expected"""

    @dataobject
    class Foo(DataObjectBase):
        foo: Union[
            Annotated[bool, FieldNumber(10), OneofField("foo_bool1")],
            Annotated[bool, FieldNumber(20), OneofField("foo_bool2")],
        ]

    # if the fields are of the same type, then by default the first one is set
    foo1 = Foo(True)
    assert foo1.which_oneof("foo") == "foo_bool1"
    assert foo1.foo_bool1
    assert foo1.foo_bool2 == None

    # unless set explicitly
    foo2 = Foo(foo_bool2=True)
    assert foo2.which_oneof("foo") == "foo_bool2"
    assert foo2.foo_bool1 == None
    assert foo2.foo_bool2


@pytest.mark.skip(
    reason="Currently, we do not have a way to use static BoolSequence in caikit core. See Issue #409"
)
def test_dataobject_with_same_type_of_oneof_with_lists():
    """Make sure that using a Union of lists to create a oneof with the same types works as expected"""

    @dataobject
    class Foo(DataObjectBase):
        foo: Union[
            Annotated[List[bool], FieldNumber(10), OneofField("foo_bool1")],
            Annotated[List[bool], FieldNumber(20), OneofField("foo_bool2")],
        ]

    # if the fields are of the same type, then by default the first one is set
    foo1 = Foo([True])
    assert foo1.which_oneof("foo") == "foo_bool1"
    assert foo1.foo_bool1[0]
    assert foo1.foo_bool2 == None
    assert Foo.from_json(foo1.to_json) == foo1

    # unless set explicitly
    foo2 = Foo(foo_bool2=[True])
    assert foo2.which_oneof("foo") == "foo_bool2"
    assert foo2.foo_bool1 == None
    assert foo2.foo_bool2[0]
    assert Foo.from_json(foo2.to_json) == foo2


def test_dataobject_primitive_oneof_round_trips():
    @dataobject
    class Foo(DataObjectBase):
        foo: Union[
            Annotated[int, FieldNumber(10), OneofField("myfooint")],
            Annotated[float, FieldNumber(20), OneofField("myfoofloat")],
        ]

    # proto round trip
    foo1 = Foo(2)
    assert foo1.which_oneof("foo") == "myfooint"
    proto_repr_foo = foo1.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    foo2 = Foo(foo=2)
    assert foo2.which_oneof("foo") == "myfooint"
    proto_repr_foo = foo2.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    foo3 = Foo(1.2)
    assert foo3.which_oneof("foo") == "myfoofloat"
    proto_repr_foo = foo3.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    foo4 = Foo(myfooint=4)
    assert foo4.which_oneof("foo") == "myfooint"
    proto_repr_foo = foo4.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # dict test
    assert foo1.to_dict() == {"myfooint": 2}
    assert foo2.to_dict() == {"myfooint": 2}
    assert foo3.to_dict() == {"myfoofloat": 1.2}
    assert foo4.to_dict() == {"myfooint": 4}

    # json round trip
    json_repr_foo = foo1.to_json()
    assert json.loads(json_repr_foo) == {
        "myfooint": 2,
    }
    assert Foo.from_json(json_repr_foo) == foo1


def test_dataobject_primitive_oneof_round_trips_no_custom_name():
    @dataobject
    class Foo(DataObjectBase):
        foo: Union[
            Annotated[int, FieldNumber(10)],
            Annotated[float, FieldNumber(20)],
        ]

    # proto round trip for Foo with no param name
    foo1 = Foo(2)
    assert foo1.which_oneof("foo") == "foo_int"
    proto_repr_foo = foo1.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with oneof name
    foo2 = Foo(foo=2)
    assert foo2.which_oneof("foo") == "foo_int"
    proto_repr_foo = foo2.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with no param name
    foo3 = Foo(1.2)
    assert foo3.which_oneof("foo") == "foo_float"
    proto_repr_foo = foo3.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with field name
    foo4 = Foo(foo_int=4)
    assert foo4.which_oneof("foo") == "foo_int"
    proto_repr_foo = foo4.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # dict test
    assert foo1.to_dict() == {"foo_int": 2}
    assert foo2.to_dict() == {"foo_int": 2}
    assert foo3.to_dict() == {"foo_float": 1.2}
    assert foo4.to_dict() == {"foo_int": 4}

    # json round trip
    json_repr_foo = foo1.to_json()
    assert json.loads(json_repr_foo) == {
        "foo_int": 2,
    }
    assert Foo.from_json(json_repr_foo) == foo1


def test_dataobject_list_oneof_round_trips_with_specified_names():
    @dataobject
    class Foo(DataObjectBase):
        foo: Union[
            Annotated[List[int], FieldNumber(10), OneofField("myfoointseq")],
            Annotated[List[str], FieldNumber(20), OneofField("myfoostrseq")],
        ]

    # proto round trip for Foo with no param name
    foo1 = Foo([2])
    assert foo1.which_oneof("foo") == "myfoointseq"
    proto_repr_foo = foo1.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with oneof param name
    foo2 = Foo(foo=[2])
    assert foo2.which_oneof("foo") == "myfoointseq"
    proto_repr_foo = foo2.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with no param name
    foo3 = Foo(["hello"])
    assert foo3.which_oneof("foo") == "myfoostrseq"
    proto_repr_foo = foo3.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with field name
    foo4 = Foo(myfoostrseq=["hello"])
    assert foo4.which_oneof("foo") == "myfoostrseq"
    proto_repr_foo = foo4.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # dict test
    assert foo1.to_dict() == {"myfoointseq": {"values": [2]}}
    assert foo2.to_dict() == {"myfoointseq": {"values": [2]}}
    assert foo3.to_dict() == {"myfoostrseq": {"values": ["hello"]}}
    assert foo4.to_dict() == {"myfoostrseq": {"values": ["hello"]}}

    # json round trip
    json_repr_foo = foo1.to_json()
    assert json.loads(json_repr_foo) == {
        "myfoointseq": {"values": [2]},
    }

    assert Foo.from_json(json_repr_foo) == foo1


def test_dataobject_list_oneof_round_trips_with_no_custom_names():
    @dataobject
    class Foo(DataObjectBase):
        foo: Union[
            Annotated[List[int], FieldNumber(10)],
            Annotated[List[str], FieldNumber(20)],
        ]

    # proto round trip for Foo with no param name
    foo1 = Foo([2])
    assert foo1.which_oneof("foo") == "foo_int_sequence"
    proto_repr_foo = foo1.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with oneof name
    foo2 = Foo(foo=[2])
    assert foo2.which_oneof("foo") == "foo_int_sequence"
    proto_repr_foo = foo2.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with no param name
    foo3 = Foo(foo=["hello"])
    assert foo3.which_oneof("foo") == "foo_str_sequence"
    proto_repr_foo = foo3.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # proto round trip for Foo with field name
    foo4 = Foo(foo_int_sequence=[2])
    assert foo4.which_oneof("foo") == "foo_int_sequence"
    proto_repr_foo = foo4.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # dict test
    assert foo1.to_dict() == {"foo_int_sequence": {"values": [2]}}
    assert foo2.to_dict() == {"foo_int_sequence": {"values": [2]}}
    assert foo3.to_dict() == {"foo_str_sequence": {"values": ["hello"]}}
    assert foo4.to_dict() == {"foo_int_sequence": {"values": [2]}}

    # json round trip
    json_repr_foo = foo1.to_json()
    assert json.loads(json_repr_foo) == {
        "foo_int_sequence": {
            "values": [2],
        },
    }
    assert Foo.from_json(json_repr_foo) == foo1


def test_dataobject_oneof_from_backend():
    """Make sure that a oneof can be correctly accessed from a backend"""

    @dataobject
    class Foo(DataObjectBase):
        foo: Union[int, str]

    data_dict1 = {"foo_int": 1234}
    backend1 = DictBackend(data_dict1)
    msg1 = Foo.from_backend(backend1)
    assert msg1.foo == 1234
    assert msg1.foo_int == 1234
    assert msg1.which_oneof("foo") == "foo_int"

    data_dict2 = {"foo": 1234}
    backend2 = DictBackend(data_dict2)
    msg2 = Foo.from_backend(backend2)
    assert msg2.foo == 1234
    assert msg2.foo_int == 1234
    assert msg2.which_oneof("foo") == "foo_int"


def test_dataobject_oneof_numeric_type_precedence():
    """Make sure that when inferring the which field from the python type, the
    value of bool < int < float is respected
    """

    @dataobject
    class Foo(DataObjectBase):
        value: Union[
            # NOTE: The order matters here. Since float is declared first, it
            #   would naturally occur before int without correct sorting
            Annotated[float, OneofField("float_val")],
            Annotated[int, OneofField("int_val")],
            Annotated[bool, OneofField("bool_val")],
        ]

    foo_bool = Foo(True)
    assert foo_bool.which_oneof("value") == "bool_val"
    foo_int = Foo(123)
    assert foo_int.which_oneof("value") == "int_val"
    foo_float = Foo(1.23)
    assert foo_float.which_oneof("value") == "float_val"


def test_dataobject_round_trip_json():
    """Make sure that a dataobject class can serialize to/from json"""

    @dataobject
    class BazObj(DataObjectBase):
        foo: str
        bar: int

    baz1 = BazObj(foo="foo", bar=1)
    js_repr = baz1.to_json()
    assert json.loads(js_repr) == {"foo": "foo", "bar": 1}
    baz2 = BazObj.from_json(js_repr)
    assert baz2.to_json() == js_repr


def test_dataobject_round_trip_proto():
    """Make sure that a dataobject class can serialize to/from proto"""

    @dataobject
    class BazObj(DataObjectBase):
        foo: str
        bar: int

    baz1 = BazObj(foo="foo", bar=1)
    proto_repr = baz1.to_proto()
    assert proto_repr.foo == "foo"
    assert proto_repr.bar == 1
    baz2 = BazObj.from_proto(proto_repr)
    assert baz2.to_proto() == proto_repr


def test_dir_on_instance():
    """This addresses a bug in how the @dataobject decorator binds in the
    DataBase base class which caused dir(x) to fail on instances of wrapped
    classes.
    """

    @dataobject
    class BazObj(DataObjectBase):
        foo: str

    x = BazObj("foobar")
    dir(x)


def test_dataobject_invocation_flavors():
    """Make sure invoking dataobject works in all of the different correct
    invocation styles and errors with all invalid styles

    VALID:
    1. No function call
    2. Function call with no args
    3. Function call with single positional argument
    4. Function call with keyword args

    INVALID:
    1. Unexpected kwargs
    2. Package as position and keyword arg
    3. No inheritance from DataObjectBase
    """
    ## Valid ##

    # 1. No function call
    @dataobject
    class Foo1(DataObjectBase):
        foo: int

    assert "foo" in Foo1._proto_class.DESCRIPTOR.fields_by_name

    # 2. Function call with no args
    @dataobject()
    class Foo2(DataObjectBase):
        foo: int

    assert "foo" in Foo2._proto_class.DESCRIPTOR.fields_by_name

    # 3. Function call with single positional argument
    @dataobject("foo.bar")
    class Foo3(DataObjectBase):
        foo: int

    assert "foo" in Foo3._proto_class.DESCRIPTOR.fields_by_name
    assert Foo3._proto_class.DESCRIPTOR.file.package == "foo.bar"

    # 4. Function call with keyword args
    @dataobject(package="foo.bar")
    class Foo4(DataObjectBase):
        foo: int

    assert "foo" in Foo4._proto_class.DESCRIPTOR.fields_by_name
    assert Foo4._proto_class.DESCRIPTOR.file.package == "foo.bar"

    ## INVALID ##

    # 1. Unexpected kwargs
    with pytest.raises(TypeError):

        @dataobject(buz="baz", package="foo.bar")
        class FooBad(DataObjectBase):
            foo: int

    # 2. Package as position and keyword arg
    with pytest.raises(TypeError):

        @dataobject("baz.bat", package="foo.bar")
        class Foo4(DataObjectBase):
            foo: int

    # 3. No inheritance from DataObjectBase
    with pytest.raises(ValueError):

        @dataobject
        class Foo4:
            foo: int


def test_dataobject_pre_existing_dataclass():
    """Make sure that wrapping a class that's already a dataclass works as
    expected by adding additional None defaults and re-making the dataclass
    """

    @dataobject
    @dataclass
    class Foo(DataObjectBase):
        foo: int

    assert is_dataclass(Foo)

    # Make sure defaults are added correctly per dataobject semantics
    inst = Foo()
    assert inst.foo is None


def test_dataobject_dataclass_non_default_init():
    """Make sure that a dataclass with a non-default __init__ does not get
    overwritten
    """

    @dataobject
    class Foo(DataObjectBase):
        foo: int

        def __init__(self, foo_base):
            self.foo = foo_base + 1

    assert is_dataclass(Foo)

    # Make sure defaults are added correctly per dataobject semantics
    inst = Foo(1)
    assert inst.foo == 2


def test_dataobject_dataclass_default_factory():
    """Make sure that a dataclass's datafactory field is preserved"""

    @dataobject
    class Foo(DataObjectBase):
        foo: List[int] = field(default_factory=list)

    assert is_dataclass(Foo)

    # Make sure default construction users the default factory
    inst = Foo()
    assert inst.foo is not None
    assert inst.foo == []


def test_enum_value_dereference():
    """Make sure that enum value objects can be used to instantiate data model
    objects and that they are correctly dereferenced in to_proto, to_dict, and
    to_json
    """

    @dataobject
    class FooEnum(Enum):
        FOO = 1
        BAR = 2

    @dataobject
    class Foo(DataObjectBase):
        foo: FooEnum

    # Create an instance with an enum value object
    inst = Foo(foo=FooEnum.FOO)
    assert inst.to_proto().foo == FooEnum.FOO.value
    assert inst.to_dict()["foo"] == FooEnum.FOO.name
    assert json.loads(inst.to_json())["foo"] == FooEnum.FOO.name

    # Create an instance with an integer value
    inst = Foo(foo=FooEnum.FOO.value)
    assert inst.to_proto().foo == FooEnum.FOO.value
    assert inst.to_dict()["foo"] == FooEnum.FOO.name
    assert json.loads(inst.to_json())["foo"] == FooEnum.FOO.name


def test_np_dtypes():
    """Make sure that numpy dtype types can be used in dataobjects"""

    @dataobject
    class Foo(DataObjectBase):
        int32: np.int32
        int64: np.int64
        uint32: np.uint32
        uint64: np.uint64
        float32: np.float32
        float64: np.float64

    descriptor = Foo._proto_class.DESCRIPTOR
    assert (
        descriptor.fields_by_name["int32"].type
        == _descriptor.FieldDescriptor.TYPE_INT32
    )
    assert (
        descriptor.fields_by_name["int64"].type
        == _descriptor.FieldDescriptor.TYPE_INT64
    )
    assert (
        descriptor.fields_by_name["uint32"].type
        == _descriptor.FieldDescriptor.TYPE_UINT32
    )
    assert (
        descriptor.fields_by_name["uint64"].type
        == _descriptor.FieldDescriptor.TYPE_UINT64
    )
    assert (
        descriptor.fields_by_name["float32"].type
        == _descriptor.FieldDescriptor.TYPE_FLOAT
    )
    assert (
        descriptor.fields_by_name["float64"].type
        == _descriptor.FieldDescriptor.TYPE_DOUBLE
    )


@pytest.mark.parametrize("run_num", range(100))
def test_dataobject_jsondict(auto_temp_dpool, run_num):
    """Make sure that a JsonDict type is handled correctly in a dataobject

    NOTE: This test is repeated 100x due to a strange segfault in `upb` that it
        can trigger. The workaround above in `temp_dpool` should solve it, but
        we retain the repetition to catch anything that's missed.
    """

    @dataobject
    class Foo(DataObjectBase):
        js_dict: JsonDict

    # Make sure the field has the right type
    Struct = auto_temp_dpool.FindMessageTypeByName("google.protobuf.Struct")
    assert Foo._proto_class.DESCRIPTOR.fields_by_name["js_dict"].message_type == Struct

    # Make sure dict is preserved on init
    js_dict = {"foo": {"bar": [1, 2, 3]}}
    foo = Foo(js_dict)
    assert foo.js_dict == js_dict

    # Make sure conversion to struct happens on to_proto
    foo_proto = foo.to_proto()
    assert set(foo_proto.js_dict.fields.keys()) == set(js_dict.keys())
    assert foo_proto.js_dict.fields["foo"].struct_value
    assert set(foo_proto.js_dict.fields["foo"].struct_value.fields.keys()) == set(
        js_dict["foo"].keys()
    )

    # Make sure conversion back to dict happens on from_proto
    foo2 = Foo.from_proto(foo_proto)
    assert foo2.js_dict == foo.js_dict


def test_dataobject_jsondict_repeated(auto_temp_dpool):
    """Make sure that a list of JsonDict types is handled correctly in a dataobject"""

    @dataobject
    class Foo(DataObjectBase):
        js_dict: List[JsonDict]

    # Make sure the field has the right type
    Struct = auto_temp_dpool.FindMessageTypeByName("google.protobuf.Struct")
    assert Foo._proto_class.DESCRIPTOR.fields_by_name["js_dict"].message_type == Struct

    # Make sure dict is preserved on init
    js_dict = [{"foo": {"bar": [1, 2, 3]}}]
    foo = Foo(js_dict)
    assert foo.js_dict == js_dict

    # Make sure conversion to struct happens on to_proto
    foo_proto = foo.to_proto()
    assert len(foo_proto.js_dict) == 1
    assert set(foo_proto.js_dict[0].fields.keys()) == set(js_dict[0].keys())
    assert foo_proto.js_dict[0].fields["foo"].struct_value
    assert set(foo_proto.js_dict[0].fields["foo"].struct_value.fields.keys()) == set(
        js_dict[0]["foo"].keys()
    )

    # Make sure conversion back to dict happens on from_proto
    foo2 = Foo.from_proto(foo_proto)
    assert foo2.js_dict == foo.js_dict


def test_dataobject_to_kwargs():
    """to_kwargs does a non-recursive version of to_dict"""

    @dataobject
    class Bar(DataObjectBase):
        bar: int

    @dataobject
    class Foo(DataObjectBase):
        int_val: int
        type_union_val: Union[int, str]
        bar_val: Bar
        type_union_list_val: Union[List[str], str]

    bar = Bar(bar=42)
    foo = Foo(
        int_val=1,
        type_union_val="foo",
        bar_val=bar,
        type_union_list_val=["hello", "world"],
    )

    kwargs = foo.to_kwargs()

    # `type_union_val` is set here rather than the `type_union_val_str_val` internal oneof field name
    assert kwargs == {
        "int_val": 1,
        "type_union_val": "foo",
        "bar_val": bar,
        "type_union_list_val": ["hello", "world"],
    }


def test_dataobject_inheritance():
    """Make sure that dataobject classes can inherit from each other in the same
    way that dataclasses can
    """

    @dataobject
    class Base(DataObjectBase):
        foo: int
        bar: int

    @dataobject
    class Derived(Base):
        bar: str
        baz: str

    # Validate Base
    desc = Base.get_proto_class().DESCRIPTOR
    fld = desc.fields_by_name["foo"]  # To save typing
    assert "foo" in desc.fields_by_name
    assert "bar" in desc.fields_by_name
    assert "baz" not in desc.fields_by_name
    assert desc.fields_by_name["foo"].type == fld.TYPE_INT64
    assert desc.fields_by_name["bar"].type == fld.TYPE_INT64
    inst = Derived(1, 2)
    assert inst.foo == 1
    assert inst.bar == 2

    # Validate Derived
    desc = Derived.get_proto_class().DESCRIPTOR
    fld = desc.fields_by_name["foo"]  # To save typing
    assert "foo" in desc.fields_by_name
    assert "bar" in desc.fields_by_name
    assert "baz" in desc.fields_by_name
    assert desc.fields_by_name["foo"].type == fld.TYPE_INT64
    assert desc.fields_by_name["bar"].type == fld.TYPE_STRING
    assert desc.fields_by_name["baz"].type == fld.TYPE_STRING
    inst = Derived(1, "asdf", "qwer")
    assert inst.foo == 1
    assert inst.bar == "asdf"
    assert inst.baz == "qwer"


def test_dataobject_union_repeated():
    """Make sure that a oneof with lists of primitive fields works correctly"""

    @dataobject
    class Foo(DataObjectBase):
        foo: Union[List[str], List[int]]
        bar: Union[List[str], List[int]]

    # The above behaves _almost_ the same way as this
    # with some naming caveats for one-of fields being
    # foo_foointsequence instead of foo_int_sequence and
    # bar_barintsequence instead of bar_int_sequence
    #
    # @dataobject
    # class Foo(DataObjectBase):
    #     @dataobject
    #     class FooIntSequence(DataObjectBase):
    #         values: List[int]
    #
    #     @dataobject
    #     class FooStrSequence(DataObjectBase):
    #         values: List[str]
    #
    #     @dataobject
    #     class BarIntSequence(DataObjectBase):
    #         values: List[int]
    #
    #     @dataobject
    #     class BarStrSequence(DataObjectBase):
    #         values: List[str]
    #
    #     foo: Union[FooIntSequence, FooStrSequence]
    #     bar: Union[BarIntSequence, BarStrSequence]

    # Foo
    # proto round trip
    foo_int = Foo.FooIntSequence(values=[1, 2])
    foo1 = Foo(foo=foo_int)
    assert foo1.which_oneof("foo") == "foo_int_sequence"
    proto_repr_foo = foo1.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # test native python list
    foo_int_native = Foo([1, 2])
    assert foo_int_native.which_oneof("foo") == "foo_int_sequence"
    proto_repr_foo = foo_int_native.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo).to_proto() == proto_repr_foo

    # dict test
    assert foo1.to_dict() == {"foo_int_sequence": {"values": [1, 2]}}
    assert foo_int_native.to_dict() == {"foo_int_sequence": {"values": [1, 2]}}

    # json round trip
    json_repr_foo = foo1.to_json()
    assert json.loads(json_repr_foo) == {"foo_int_sequence": {"values": [1, 2]}}
    foo_json_repr = Foo.from_json(json_repr_foo)
    assert foo_json_repr.to_json() == json_repr_foo

    # test native python list json round trip
    json_repr_foo = foo_int_native.to_json()
    assert json.loads(json_repr_foo) == {"foo_int_sequence": {"values": [1, 2]}}
    foo_json_repr = Foo.from_json(json_repr_foo)
    assert foo_json_repr.to_json() == json_repr_foo

    foo_str = Foo.FooStrSequence(values=["hello", "world"])
    foo2 = Foo(foo=foo_str)
    assert foo2.which_oneof("foo") == "foo_str_sequence"
    proto_repr_foo2 = foo2.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo2).to_proto() == proto_repr_foo2

    # test native python list
    foo_str_native = Foo(["hello", "world"])
    assert foo_str_native.which_oneof("foo") == "foo_str_sequence"
    proto_repr_foo2 = foo_str_native.to_proto()
    assert Foo.from_proto(proto=proto_repr_foo2).to_proto() == proto_repr_foo2

    # dict test
    assert foo2.to_dict() == {"foo_str_sequence": {"values": ["hello", "world"]}}
    assert foo_str_native.to_dict() == {
        "foo_str_sequence": {"values": ["hello", "world"]}
    }

    # Bar
    # proto round trip
    bar_int = Foo.BarIntSequence(values=[1, 2])
    bar1 = Foo(bar=bar_int)
    assert bar1.which_oneof("bar") == "bar_int_sequence"
    proto_repr_bar = bar1.to_proto()
    assert Foo.from_proto(proto=proto_repr_bar).to_proto() == proto_repr_bar

    # dict test
    assert bar1.to_dict() == {"bar_int_sequence": {"values": [1, 2]}}

    # json round trip
    json_repr_bar = bar1.to_json()
    assert json.loads(json_repr_bar) == {"bar_int_sequence": {"values": [1, 2]}}
    bar_json_repr = Foo.from_json(json_repr_bar)
    assert bar_json_repr.to_json() == json_repr_bar

    bar_str = Foo.BarStrSequence(values=["hello", "world"])
    bar2 = Foo(bar=bar_str)
    assert bar2.which_oneof("bar") == "bar_str_sequence"
    proto_repr_bar2 = bar2.to_proto()
    assert Foo.from_proto(proto=proto_repr_bar2).to_proto() == proto_repr_bar2


def test_dataobject_function_inheritance():
    """Make sure inheritance works to override functionality without changing
    the schema of the parent
    """

    @dataobject
    class Base(DataObjectBase):
        foo: int

        def doit(self):
            return self.foo * 2

    @dataobject
    class Derived(Base):
        def doit(self):
            return self.foo * 3

    b_inst = Base(1)
    assert b_inst.doit() == 2

    d_inst = Derived(1)
    assert d_inst.doit() == 3


def test_make_dataobject_no_optionals():
    """Test that dataobject classes can be created dynamically without optional
    values given
    """
    data_object = make_dataobject(
        name="FooBar",
        annotations={"foo": str, "bar": int},
    )

    assert data_object.__name__ == "FooBar"
    assert data_object.__annotations__ == {"foo": str, "bar": int}
    assert issubclass(data_object, DataObjectBase)
    assert data_object._proto_class.DESCRIPTOR.name == "FooBar"
    assert data_object._proto_class.DESCRIPTOR.file.package == CAIKIT_DATA_MODEL


def test_make_dataobject_with_optionals():
    """Test that dataobject classes can be created dynamically with optional
    values given
    """

    class OtherBase:
        pass

    data_object = make_dataobject(
        name="FooBar",
        annotations={"foo": str, "bar": int},
        bases=(OtherBase,),
        attrs={"prop": "val"},
        proto_name="SomeOtherFooBar",
        package="foo.bar.baz",
    )

    assert data_object.__name__ == "FooBar"
    assert data_object.__annotations__ == {"foo": str, "bar": int}
    assert issubclass(data_object, DataObjectBase)
    assert issubclass(data_object, OtherBase)
    assert data_object.prop == "val"
    assert data_object._proto_class.DESCRIPTOR.name == "SomeOtherFooBar"
    assert data_object._proto_class.DESCRIPTOR.file.package == "foo.bar.baz"


def test_dataobject_repr_data_base():
    """Make sure that the __repr__ on a dataobject uses DataBase.__repr__"""

    @dataobject
    class Foo(DataObjectBase):
        foo: int

    inst = Foo(1)
    assert repr(inst) == DataBase.__repr__(inst)


def test_dataobject_custom_repr():
    """Make sure that a dataobject with a custom __repr__ retains it"""
    custom_repr = "so custom!"

    @dataobject
    class Foo(DataObjectBase):
        foo: int

        def __repr__(self):
            return custom_repr

    inst = Foo(1)
    assert repr(inst) == custom_repr



# Standard
import os

# Local
from . import data_model, modules
from .modules import (
    CompositeModule,
    InnerModule,
    OtherModule,
    SampleModule,
    SamplePrimitiveModule,
    StreamingModule,
)
from caikit.config import configure

# Run configure for sample_lib configuration
configure(os.path.join(os.path.dirname(__file__), "config.yml"))



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Single spot for all global shared state access

This should not import anything else from caikit outside the toolkit error handler, many parts of
caikit need to access this state so that would easily cause an import cycle.
"""

# Standard
from typing import Any, Dict, Tuple, Type

# First Party
import alog

# Local
from .exceptions import error_handler

log = alog.use_channel("REGISTRIES")
error = error_handler.get(log)


# Registry of all classes decorated with `@caikit.module`
MODULE_REGISTRY = {}

# Global module backend registry dict
MODULE_BACKEND_REGISTRY = {}


def module_registry() -> Dict[str, "caikit.core.ModuleBase"]:
    """🌶️🌶️🌶️ This returns global state that should only be mutated if you know what you're doing!

    Returns the dictionary of decorated @modules that have been imported.
    Used to map module IDs to the concrete class implementations to load and run.

    Structure is
    Dict[ module_id, module_class ]

    Returns:
        Dict[str, caikit.core.ModuleBase]: The module registry
    """
    return MODULE_REGISTRY


def module_backend_registry() -> Dict[
    str, Dict[str, Tuple["caikit.core.ModuleBase", Dict]]
]:
    """🌶️🌶️🌶️ This returns global state that should only be mutated if you know what you're doing!

    Returns the module backend registry. This adds more nesting to the module registry,
    providing a dictionary of backend type name -> (backend module impl class, config dict)

    Structure is
    Dict[ module_id, Dict[ backend_type, Tuple[ backend_impl_class, backend_config_dict ] ] ]

    Returns:
        Dict[str, Dict[str, Tuple["caikit.core.ModuleBase", Dict]]]: The module
            backend registry
    """
    # TODO: put a real data structure here instead of nested dicts
    return MODULE_BACKEND_REGISTRY


class _AttrAccessBackendDict(dict):
    """Simple extension on a dict that allows attribute access in addition to
    index lookup
    """

    def __getattr__(self, name: str) -> Any:
        """Alias to index lookup"""
        error.value_check(
            "<COR85015051E>", name in self, "backend type {} not registered", name
        )
        return self[name]


# "enum" holding known backend types. This is implemented as a dict so that it
# can be extended as needed by downstream libraries.
MODULE_BACKEND_TYPES = _AttrAccessBackendDict()


def module_backend_types() -> Dict[str, str]:
    """🌶️🌶️🌶️ This returns global state that should only be mutated if you know what you're doing!

    Returns the "enum" of module backend types. This is a dict where the keys and values
    are identical, and each are the string names of a backend type.

    Returns:
        Dict[str, str]:
            The module backend type enum
    """
    return MODULE_BACKEND_TYPES


MODULE_BACKEND_CLASSES: Dict[str, Type["caikit.core.BackendBase"]] = {}


def module_backend_classes() -> Dict[str, Type["caikit.core.BackendBase"]]:
    """🌶️🌶️🌶️ This returns global state that should only be mutated if you know what you're doing!

    Returns the mapping of backend type name to concrete backend class

    Returns:
        Dict[str, Type["caikit.core.BackendBase"]]:
            The module backend class registry
    """

    return MODULE_BACKEND_CLASSES



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A ModelFinder is responsible for finding models based on a unique input string
which may be a path, id, or other identifier.

Configuration for ModelFinders lives under the config as follows:

model_management:
    finders:
        <finder name>:
            type: <finder type name>
            config:
                <config option>: <value>
"""

# Standard
from typing import Optional
import abc

# Local
from ..modules import ModuleConfig
from ..toolkit.factory import FactoryConstructible


class ModelFinderBase(FactoryConstructible):
    __doc__ = __doc__

    @abc.abstractmethod
    def find_model(
        self,
        model_path: str,
        **kwargs,
    ) -> Optional[ModuleConfig]:
        """Find any model that can be uniquely identified by the given (logical)
        path. If found, return the in-memory ModuleConfig.

        Args:
            model_path (str): The logical path to the model (name, id, file
                path, etc...)
            **kwargs: All finders must allow additional kwargs through so that
                specific finders and loaders can support additional optional
                arguments.

        Returns:
            result (Optional[ModuleConfig]): If found, the in-memory config
                object for the model. If the model cannot be found, an exception
                may be raised or None returned.
        """



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit.core.data_model import DataStream
from caikit.core.data_model.streams.csv_column_formatter import CSVColumnFormatter

# Unit Test Infrastructure
from tests.base import TestCaseBase


class TestCSVColumnFormatter(TestCaseBase):
    def test_does_nothing_if_no_expected_types_are_lists(self):
        formatter = CSVColumnFormatter({"a": int, "b": int})
        list_stream = DataStream.from_iterable([[1, 2], [3, 4]])
        self.assertEqual(list_stream, formatter.format(list_stream))

    def test_it_is_okay_with_empty_streams(self):
        formatter = CSVColumnFormatter({"a": int, "b": list})
        empty_stream = DataStream.from_iterable([])
        stream = formatter.format(empty_stream)
        self.assertEqual([], list(stream))

    def test_it_listifies_basic_types_on_list_streams(self):
        formatter = CSVColumnFormatter({"a": int, "b": list})
        basic_stream = [[1, 2], [3, 4.0], [5, "6.0"], [7, True], [8, [9]]]
        expected_stream = [[1, [2]], [3, [4.0]], [5, ["6.0"]], [7, [True]], [8, [9]]]
        stream = DataStream.from_iterable(basic_stream)
        self.assertEqual(expected_stream, list(formatter.format(stream)))

    def test_it_does_not_listify_dicts_or_tuples(self):
        formatter = CSVColumnFormatter({"a": list})
        # Dicts and Tuples as elements of a list stream are not converted
        stream = [[{"foo": "bar"}], [("foo", "bar")]]
        stream = DataStream.from_iterable(stream)
        self.assertEqual(list(stream), list(formatter.format(stream)))

    def test_it_collapses_extra_elements_into_a_list(self):
        formatter = CSVColumnFormatter({"text": str, "labels": list})
        stream = [["foo", "bar"], ["foo", "bar", "baz", "buz"]]
        expected_stream = [["foo", ["bar"]], ["foo", ["bar", "baz", "buz"]]]
        stream = DataStream.from_iterable(stream)
        self.assertEqual(expected_stream, list(formatter.format(stream)))

    def test_it_does_not_process_items_that_are_not_lists(self):
        formatter = CSVColumnFormatter({"text": str, "labels": list})
        stream = [["foo", "bar"], "baz", 5, ("foo", "bar"), {"foo": "bar"}]
        expected_stream = [["foo", ["bar"]], "baz", 5, ("foo", "bar"), {"foo": "bar"}]
        stream = DataStream.from_iterable(stream)
        self.assertEqual(expected_stream, list(formatter.format(stream)))

    def test_it_does_not_barf_if_lists_have_fewer_elements_than_expected(self):
        formatter = CSVColumnFormatter({"text": str, "labels": list})
        stream = [["foo", "bar"], ["baz"], [5]]
        expected_stream = [["foo", ["bar"]], ["baz"], [5]]
        stream = DataStream.from_iterable(stream)
        self.assertEqual(expected_stream, list(formatter.format(stream)))



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A Trainer is responsible for managing execution of a training job for a given
module class

Configuration for ModelTrainers lives under the config as follows:

model_management:
    trainers:
        <trainer name>:
            type: <trainer type name>
            config:
                <config option>: <value>
"""
# Standard
from typing import List, Optional, Type, Union
import abc
import dataclasses
import datetime
import os

# Local
from ...interfaces.common.data_model.stream_sources import S3Path
from ..data_model import TrainingStatus
from ..modules import ModuleBase
from ..toolkit.factory import FactoryConstructible
from ..toolkit.reversible_hasher import ReversibleHasher


@dataclasses.dataclass
class TrainingInfo:
    status: TrainingStatus
    errors: Optional[List[Exception]] = None
    submission_time: Optional[datetime.datetime] = None
    completion_time: Optional[datetime.datetime] = None


class ModelTrainerBase(FactoryConstructible):
    __doc__ = __doc__

    class ModelFutureBase(abc.ABC):
        """Every Trainer must implement a ModelFuture class that can access the
        training job in the infrastructure managed by the trainer.
        """

        ID_DELIMITER = ":"

        def __init__(
            self,
            trainer_name: str,
            training_id: str,
            save_with_id: bool,
            save_path: Optional[Union[str, S3Path]],
            model_name: Optional[str] = None,
            use_reversible_hash: bool = True,
        ):
            # Trainers should deal with an S3 ref first and not pass it along here
            if save_path and isinstance(save_path, S3Path):
                raise ValueError("S3 output path not supported by this runtime")
            self._id = (
                self.__class__.ID_DELIMITER.join(
                    [ReversibleHasher.hash(trainer_name), training_id]
                )
                if use_reversible_hash
                else training_id
            )
            self._save_path = self.__class__._save_path_with_id(
                save_path,
                save_with_id,
                self._id,
                model_name,
            )

        @property
        def id(self) -> str:
            """Every model future must have a unique ID that can be used to look
            up the in-flight training
            """
            return self._id

        @property
        def save_path(self) -> Optional[str]:
            """If created with a save path, the future must expose it, including
            any injected training id
            """
            return self._save_path

        @abc.abstractmethod
        def get_info(self) -> TrainingInfo:
            """Every model future must be able to poll the status of the
            training job
            """

        @abc.abstractmethod
        def cancel(self):
            """Terminate the given training"""

        @abc.abstractmethod
        def wait(self):
            """Block until the job reaches a terminal state"""

        @abc.abstractmethod
        def load(self) -> ModuleBase:
            """A model future must be loadable with no additional arguments"""

        ## Common Impl ##

        def result(self) -> ModuleBase:
            """Support result() to match concurrent.futures.Future"""
            return self.load()

        @classmethod
        def _save_path_with_id(
            cls,
            save_path: Optional[str],
            save_with_id: bool,
            training_id: str,
            model_name: Optional[str],
        ) -> Optional[str]:
            """If asked to save_with_id, child classes should use this shared
            utility to construct the final save path
            """
            if save_path is None:
                return save_path

            final_path_parts = [save_path]
            # If told to save with the ID in the path, inject it before the
            # model name.
            if save_with_id and training_id not in save_path:
                # (Don't inject training id if its already in the path)
                final_path_parts.append(training_id)

            if model_name and model_name not in save_path:
                final_path_parts.append(model_name)

            return os.path.join(*final_path_parts)

    @abc.abstractmethod
    def train(
        self,
        module_class: Type[ModuleBase],
        *args,
        save_path: Optional[Union[str, S3Path]] = None,
        save_with_id: bool = False,
        model_name: Optional[str] = None,
        **kwargs,
    ) -> "ModelFutureBase":
        """Start training the given module and return a future to the trained
        model instance
        """

    @abc.abstractmethod
    def get_model_future(self, training_id: str) -> "ModelFutureBase":
        """Look up the model future for the given id"""

    ## Shared Utilities ##

    @classmethod
    def get_trainer_name(cls, training_id: str) -> str:
        """Un-hash the trainer's instance name from the given training id"""
        return ReversibleHasher.reverse_hash(
            training_id.split(cls.ModelFutureBase.ID_DELIMITER)[0]
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The LocalModelFinder locates models locally on disk that contain a caikit-native
config.yml file.

Configuration for LocalModelFinder lives under the config as follows:

model_management:
    finders:
        <finder name>:
            type: LOCAL
            config:
                # Path to a local directory that holds models for relative paths
                load_path: <null or str>
"""
# Standard
from typing import Optional
import os

# First Party
import aconfig
import alog

# Local
from ..exceptions import error_handler
from ..modules import ModuleConfig
from .model_finder_base import ModelFinderBase

log = alog.use_channel("LFIND")
error = error_handler.get(log)


class LocalModelFinder(ModelFinderBase):
    __doc__ = __doc__

    name = "LOCAL"

    def __init__(self, config: aconfig.Config, instance_name: str):
        """Initialize with an optional path prefix"""
        self._instance_name = instance_name
        self._load_path = config.load_path

    def find_model(
        self,
        model_path: str,
        **__,
    ) -> Optional[ModuleConfig]:
        """Find a model at the local path or with the configured prefix"""
        full_model_path = model_path
        if not os.path.exists(model_path) and self._load_path:
            full_model_path = os.path.join(self._load_path, model_path)
            log.debug2(
                "Looking for %s in %s using %s",
                model_path,
                full_model_path,
                self._instance_name,
            )
        full_model_path = os.path.normpath(full_model_path)
        return ModuleConfig.load(full_model_path)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Standard
import random

# First Party
import alog

# Local
from ..exceptions import error_handler

log = alog.use_channel("AUGMENT_BASE")
error = error_handler.get(log)


class AugmentorBase:
    def __init__(self, random_seed, produces_none=False):
        """The base class from which all Augmentors inherit. An augmentor is a bit different from a
        module in that it is subject to the following constraints / design considerations.

        - An augmentor must have the same input type as output type. Currently, this should only be
        data model objects / strings only.
        - An augmentor only takes one argument at run time, which is the object being augmented.
        Everything else should be contained within the class. This is necessary so that augmentors
        can be merged.

        Augmentors can only be combined if they have the same input/output object type.
        """
        error.type_check("<COR73155110E>", int, random_seed=random_seed)
        error.type_check(
            "<COR73301070E>", bool, allow_none=True, produces_none=produces_none
        )
        if random_seed is not None:
            random.seed(random_seed)
            self._is_random = True
            self._init_state = random.getstate()
        self.produces_none = produces_none

    def augment(self, inp_obj):
        """Take an object in, give an object back. Calls ._augment in the subclass.

        Args:
            inp_obj (str | caikit.core.data_model.DataBase): Object to be
                augmented.
        Returns:
            str | caikit.core.data_model.DataBase: Augmented object of same type
                as input inp_obj.
        """
        if not isinstance(inp_obj, self.augmentor_type):
            raise TypeError(
                "Input type [{}] is misaligned with augmentor type [{}]".format(
                    type(inp_obj), self.augmentor_type
                )
            )
        out_obj = self._augment(inp_obj)
        # Only some augmentors are allowed to return None.
        if out_obj is None and self.produces_none:
            return out_obj
        if not isinstance(out_obj, self.augmentor_type):
            error(
                "<COR72611702E>",
                TypeError(
                    "Output type [{}] is misaligned with augmentor type [{}]".format(
                        type(out_obj), self.augmentor_type
                    )
                ),
            )
        return out_obj

    def reset(self):
        """Reset random number generation for the current augmentor. Note that this currently
        assumes the augmentor is using the builtin random generator leveraged by Python; if
        you end up using something else, you may want to override this or restructure this
        base class to allow resetting of random states based on seed type.
        """
        if self._is_random:
            random.setstate(self._init_state)



# Standard
import signal

# First Party
import alog

# Local
from caikit.config.config import get_config

log = alog.use_channel("RUNTIME-MAIN")


def main():
    _grpc_server = None
    _http_server = None

    def interrupt(signal_, _stack_frame):
        log.info(
            "<RUN87630120I>",
            "Caikit Runtime received interrupt signal %s, shutting down",
            signal_,
        )
        if _grpc_server:
            _grpc_server.stop()
        if _http_server:
            _http_server.stop()

    # NOTE: signal function can only be called from main thread of the main
    # interpreter. If this function is called from a thread (like in tests)
    # then signal handler cannot be used. Thus, we will only have real
    # termination_handler when this is called from the __main__.

    signal.signal(signal.SIGINT, interrupt)
    signal.signal(signal.SIGTERM, interrupt)

    #####################
    # Start the servers
    #####################

    # Start serving grpc server
    if get_config().runtime.grpc.enabled:
        # Import the gRPC components inside the function to avoid requiring
        # them when starting runtime without the `runtime-grpc` optional
        # dependencies installed.

        try:
            # Local
            from caikit.runtime.grpc_server import (  # pylint: disable=import-outside-toplevel
                RuntimeGRPCServer,
            )
        except ModuleNotFoundError as e:
            message = (
                "Error: {} - unable to start gRPC server. Perhaps you missed"
                " installing the gRPC optional dependencies with"
                "`pip install caikit[runtime-grpc]`?".format(e)
            )
            log.error("<RUN72169927E>", message)
            raise

        log.debug("Starting up caikit.runtime.grpc_server")

        _grpc_server = RuntimeGRPCServer()
        _grpc_server.start(blocking=not get_config().runtime.http.enabled)

    # Start serving http server
    if get_config().runtime.http.enabled:
        # Import the HTTP components inside the function to avoid requiring
        # them when starting runtime without the `runtime-http` optional
        # dependencies installed.

        try:
            # Local
            from caikit.runtime.http_server import (  # pylint: disable=import-outside-toplevel
                RuntimeHTTPServer,
            )
        except ModuleNotFoundError as e:
            message = (
                "Error: {} - unable to start REST server. Perhaps you missed"
                " installing the http optional dependencies with"
                " `pip install caikit[runtime-http]`?".format(e)
            )
            log.error("<RUN76169927E>", message)
            raise

        log.debug("Starting up caikit.runtime.http_server")

        _http_server = RuntimeHTTPServer()
        _http_server.start(blocking=True)  # make http always blocking


if __name__ == "__main__":
    main()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from typing import Iterable
import uuid

# Local
from caikit.core import LocalBackend
from caikit.runtime.service_generation.create_service import (
    create_inference_rpcs,
    create_training_rpcs,
)
from sample_lib.data_model import (
    OtherOutputType,
    SampleInputType,
    SampleOutputType,
    SampleTask,
)
from sample_lib.modules import SampleModule
import caikit
import sample_lib

## Setup ########################################################################

widget_class = sample_lib.modules.sample_task.SampleModule

## Tests ########################################################################

### create_inference_rpcs tests #################################################

# Local
from tests.core.helpers import MockBackend


def test_create_inference_rpcs_uses_task_from_module_decorator():
    # make a new module with SampleTask
    @caikit.module(
        id=str(uuid.uuid4()), name="something", version="0.0.0", task=SampleTask
    )
    class NewModule(caikit.core.ModuleBase):
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

    # SampleModule also implements `SampleTask`
    rpcs = create_inference_rpcs([NewModule, SampleModule])
    assert len(rpcs) == 3  # SampleModule has 3 streaming flavors
    assert NewModule in rpcs[0].module_list
    assert SampleModule in rpcs[0].module_list


def test_create_inference_rpcs_includes_backend_modules():
    @caikit.module(
        id=str(uuid.uuid4()), name="something", version="0.0.0", task=SampleTask
    )
    class NewModule(caikit.core.ModuleBase):
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

    @caikit.module(backend_type=MockBackend.backend_type, base_module=NewModule)
    class NewBackendModule(caikit.core.ModuleBase):
        def run(
            self, sample_input: SampleInputType, backend_param: str
        ) -> SampleOutputType:
            pass

    rpcs = create_inference_rpcs([NewModule, NewBackendModule])
    assert len(rpcs) == 1
    assert NewModule in rpcs[0].module_list
    # Make sure the backend module and its parameters are added as well
    assert NewBackendModule in rpcs[0].module_list
    assert len(rpcs[0].request.triples) == 2
    assert rpcs[0].request.triples[1][1] == "backend_param"


def test_create_inference_rpcs_uses_task_from_module_decorator_with_streaming():
    @caikit.module(
        id=str(uuid.uuid4()),
        name="NewStreamingModule1",
        version="0.0.0",
        task=SampleTask,
    )
    class NewStreamingModule1(caikit.core.ModuleBase):
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

    @caikit.module(
        id=str(uuid.uuid4()),
        name="NewStreamingModule2",
        version="0.0.0",
        task=SampleTask,
    )
    class NewStreamingModule2(caikit.core.ModuleBase):
        @SampleTask.taskmethod()
        def run_unary(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

        @SampleTask.taskmethod(input_streaming=True, output_streaming=True)
        def run_stream_bidi(
            self, sample_inputs: caikit.core.data_model.DataStream[SampleInputType]
        ) -> caikit.core.data_model.DataStream[SampleOutputType]:
            pass

    @caikit.module(
        id=str(uuid.uuid4()),
        name="NewStreamingModule3",
        version="0.0.0",
        task=SampleTask,
    )
    class NewStreamingModule3(caikit.core.ModuleBase):
        @SampleTask.taskmethod(input_streaming=True)
        def run_stream_in(
            self, sample_inputs: caikit.core.data_model.DataStream[SampleInputType]
        ) -> SampleOutputType:
            pass

    @caikit.core.task(
        unary_parameters={"text": str},
        streaming_parameters={"texts": Iterable[str]},
        unary_output_type=OtherOutputType,
        streaming_output_type=Iterable[OtherOutputType],
    )
    class OtherStreamingTask(caikit.core.TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()),
        name="TestModule1",
        version="0.0.0",
        task=OtherStreamingTask,
    )
    class TestModule1(caikit.core.ModuleBase):
        @OtherStreamingTask.taskmethod(input_streaming=True)
        def run_stream_in(
            self, texts: caikit.core.data_model.DataStream[str]
        ) -> OtherOutputType:
            pass

        @OtherStreamingTask.taskmethod(output_streaming=True)
        def run_stream_out(
            self, text: str
        ) -> caikit.core.data_model.DataStream[OtherOutputType]:
            pass

    # Not including NewStreamingModule3 to check that we don't get ClientStreaming RPC generated
    rpcs = create_inference_rpcs(
        [NewStreamingModule1, NewStreamingModule2, SampleModule]
    )
    assert len(rpcs) == 3
    _test_rpc(
        rpcs,
        task=SampleTask,
        input_streaming=False,
        output_streaming=False,
        expected_name="SampleTaskPredict",
        expected_module_list=[NewStreamingModule1, NewStreamingModule2, SampleModule],
    )  # unary
    _test_rpc(
        rpcs,
        task=SampleTask,
        input_streaming=True,
        output_streaming=True,
        expected_name="BidiStreamingSampleTaskPredict",
        expected_module_list=[NewStreamingModule2, SampleModule],
    )  # bidi stream
    _test_rpc(
        rpcs,
        task=SampleTask,
        input_streaming=False,
        output_streaming=True,
        expected_name="ServerStreamingSampleTaskPredict",
        expected_module_list=[SampleModule],
    )  # out stream

    rpcs = create_inference_rpcs(
        [
            NewStreamingModule1,
            NewStreamingModule2,
            NewStreamingModule3,
            SampleModule,
            TestModule1,
        ]
    )
    assert len(rpcs) == 6
    # only checking the new rpcs here
    _test_rpc(
        rpcs,
        task=SampleTask,
        input_streaming=True,
        output_streaming=False,
        expected_name="ClientStreamingSampleTaskPredict",
        expected_module_list=[NewStreamingModule3],
    )  # in stream
    _test_rpc(
        rpcs,
        task=OtherStreamingTask,
        input_streaming=True,
        output_streaming=False,
        expected_name="ClientStreamingOtherStreamingTaskPredict",
        expected_module_list=[TestModule1],
    )  # OtherStreamingTask's in stream
    _test_rpc(
        rpcs,
        task=OtherStreamingTask,
        input_streaming=False,
        output_streaming=True,
        expected_name="ServerStreamingOtherStreamingTaskPredict",
        expected_module_list=[TestModule1],
    )  # OtherStreamingTask's out stream


def _test_rpc(
    rpcs, task, input_streaming, output_streaming, expected_name, expected_module_list
):
    rpc_list = [
        rpc
        for rpc in rpcs
        if rpc.task == task
        and rpc._input_streaming == input_streaming
        and rpc._output_streaming == output_streaming
    ]
    assert len(rpc_list) == 1
    rpc = rpc_list[0]
    assert rpc.name == expected_name
    assert set(rpc.module_list) == set(expected_module_list)


def test_create_inference_rpcs():
    rpcs = create_inference_rpcs([widget_class])
    assert len(rpcs) == 3  # SampleModule has inference methods for 3 streaming flavors
    assert widget_class in rpcs[0].module_list


def test_create_inference_rpcs_for_multiple_modules_of_same_type():
    module_list = [
        sample_lib.modules.sample_task.SampleModule,
        sample_lib.modules.sample_task.SamplePrimitiveModule,
        sample_lib.modules.other_task.OtherModule,
    ]
    rpcs = create_inference_rpcs(module_list)

    # 4 RPCs, SampleModule and SamplePrimitiveModule have task SampleTask with 3 flavors for
    # streaming, OtherModule has task OtherTask
    assert len(rpcs) == 4
    assert sample_lib.modules.sample_task.SampleModule in rpcs[0].module_list
    assert sample_lib.modules.sample_task.SamplePrimitiveModule in rpcs[0].module_list
    assert sample_lib.modules.sample_task.SampleModule in rpcs[1].module_list
    assert sample_lib.modules.sample_task.SampleModule in rpcs[2].module_list
    assert sample_lib.modules.other_task.OtherModule in rpcs[-1].module_list


def test_create_inference_rpcs_removes_modules_with_no_task():
    module_list = [
        sample_lib.modules.sample_task.SampleModule,  # has a task, has 3 streaming flavors
        sample_lib.modules.sample_task.InnerModule,  # does not have a task
    ]
    rpcs = create_inference_rpcs(module_list)

    assert len(rpcs) == 3
    assert sample_lib.modules.sample_task.SampleModule in rpcs[0].module_list
    assert sample_lib.modules.sample_task.InnerModule not in rpcs[0].module_list


### create_training_rpcs tests #################################################


def test_no_training_rpcs_module_with_no_train_function():
    @caikit.module(
        id=str(uuid.uuid4()), name="something", version="0.0.0", task=SampleTask
    )
    class Foo(caikit.core.ModuleBase):
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

        def train_in_progress(self):
            pass

    rpcs = create_training_rpcs([Foo])
    assert len(rpcs) == 0


def test_no_training_rpcs_for_module_with_no_task():
    @caikit.module(id=str(uuid.uuid4()), name="something", version="0.0.0")
    class Foo(caikit.core.ModuleBase):
        def train(self, foo: int) -> "Foo":
            pass

    rpcs = create_training_rpcs([Foo])
    assert len(rpcs) == 0


def test_create_training_rpcs():
    rpcs = create_training_rpcs([widget_class])
    assert len(rpcs) == 1
    assert widget_class in rpcs[0].module_list



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for backend type logic
"""

# Standard

# Third Party
import pytest

# Local
from caikit.core import backend_types
from caikit.core.module_backends.base import BackendBase
from tests.conftest import reset_backend_types
from tests.core.helpers import MockBackend

## Tests #######################################################################


def test_type_attr_forwarding():
    """Test that the LOCAL type behaves as expected"""
    assert backend_types.LOCAL == "LOCAL"


def test_register_backend_type_with_config(reset_backend_types):
    """Test that a backend type can be registered with a callable"""

    backend_types.register_backend_type(MockBackend)
    assert backend_types.MOCK == "MOCK"


def test_non_registered_backend_type_raises():
    """Test that non registered backend type raises error"""
    with pytest.raises(ValueError):
        backend_types.DUMMY


def test_register_backend_type_uppercase_required():
    """Make sure that the type name is required to be uppercase"""

    class DummyBackend(BackendBase):
        backend_type = "foo"

        def start(self):
            pass

        def register_config(self, config=...):
            pass

    with pytest.raises(ValueError):
        backend_types.register_backend_type(DummyBackend)


def test_register_backend_type_callable_required():
    """Make sure that the configuration function must be a callable"""
    with pytest.raises(TypeError):
        backend_types.register_backend_type("FOO")


def test_incorrect_backend_type_raises():
    """Test that if backend is not subclassed from BackendBase then registration
    raises error"""

    class MockBackend:
        pass

    with pytest.raises(TypeError):
        backend_types.register_backend_type(MockBackend)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DataValidationError(Exception):
    """This error is used for data validation problems during training"""

    def __init__(self, reason, item_number=None):
        if item_number:
            message = "Training data validation failed on item {}. {}".format(
                item_number, reason
            )
        else:
            message = "Training data validation failed: {}".format(reason)
        super().__init__(message)
        self._reason = reason
        self._item_number = item_number

    @property
    def reason(self) -> str:
        """The reason given for this data validation error"""
        return self._reason

    @property
    def item_number(self) -> int:
        """The index of the training data item that failed validation. Probably zero indexed"""
        return self._item_number



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import contextmanager
from functools import partial
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock, patch
import os
import shutil
import threading
import time

# Third Party
import grpc
import pytest

# Local
from caikit import get_config
from caikit.core.model_manager import ModelManager as CoreModelManager
from caikit.core.modules import ModuleBase
from caikit.runtime.model_management.loaded_model import LoadedModel
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.utils.import_util import get_dynamic_module
from tests.conftest import TempFailWrapper, random_test_id, temp_config
from tests.core.helpers import TestFinder
from tests.fixtures import Fixtures
import caikit.runtime.model_management.model_loader

get_dynamic_module("caikit.core")
ANY_MODEL_TYPE = "test-any-model-type"
ANY_MODEL_PATH = "test-any-model-path"
MODEL_MANAGER = ModelManager.get_instance()


@pytest.fixture(autouse=True)
def tear_down():
    yield
    MODEL_MANAGER.unload_all_models()


@contextmanager
def temp_local_models_dir(workdir, model_manager=MODEL_MANAGER):
    prev_local_models_dir = model_manager._local_models_dir
    model_manager._local_models_dir = workdir
    yield
    model_manager._local_models_dir = prev_local_models_dir


@contextmanager
def non_singleton_model_managers(num_mgrs=1, *args, **kwargs):
    with temp_config(*args, **kwargs):
        with patch(
            "caikit.runtime.model_management.model_loader.MODEL_MANAGER",
            new_callable=CoreModelManager,
        ):
            instances = []
            try:
                for _ in range(num_mgrs):
                    ModelManager._ModelManager__instance = None
                    instances.append(ModelManager.get_instance())
                yield instances
            finally:
                for inst in instances:
                    inst.shut_down()
                ModelManager._ModelManager__instance = MODEL_MANAGER


class SlowLoader:
    """Helper class to simulate slow loading"""

    def __init__(self, load_result="STUB"):
        self._load_result = load_result
        self._load_start_event = threading.Event()
        self._load_end_event = threading.Event()

    def load(self, *_, **__):
        self._load_start_event.wait()
        self._load_end_event.set()
        if isinstance(self._load_result, Exception):
            raise self._load_result
        return self._load_result

    def unblock_load(self):
        self._load_start_event.set()

    def done_loading(self):
        return self._load_end_event.is_set()


# ****************************** Integration Tests ****************************** #
# These tests do not patch in mocks, the manager will use real instances of its dependencies


def test_load_model_ok_response():
    model_id = "happy_load_test"
    model_size = MODEL_MANAGER.load_model(
        model_id=model_id,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert model_size > 0


@pytest.mark.parametrize("model_id", ["", b"bytes-id", 123])
def test_retrieve_invalid_model_ids(model_id):
    with pytest.raises(CaikitRuntimeException) as context:
        MODEL_MANAGER.retrieve_model(model_id)
    assert context.value.status_code == grpc.StatusCode.INVALID_ARGUMENT


def test_load_model_no_size_update():
    model_id = random_test_id()
    model_size = MODEL_MANAGER.load_model(
        model_id=model_id,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert model_size > 0
    loaded_model = MODEL_MANAGER.loaded_models[model_id]
    assert loaded_model.size() == model_size
    loaded_model.set_size(model_size * 2)
    assert loaded_model.size() == model_size


def test_load_local_models():
    with TemporaryDirectory() as tempdir:
        shutil.copytree(Fixtures.get_good_model_path(), os.path.join(tempdir, "model1"))
        shutil.copy(
            Fixtures.get_good_model_archive_path(),
            os.path.join(tempdir, "model2.zip"),
        )

        with temp_local_models_dir(tempdir):
            MODEL_MANAGER.sync_local_models(wait=True)
        assert len(MODEL_MANAGER.loaded_models) == 2
        assert "model1" in MODEL_MANAGER.loaded_models.keys()
        assert "model2.zip" in MODEL_MANAGER.loaded_models.keys()
        assert "model-does-not-exist.zip" not in MODEL_MANAGER.loaded_models.keys()


def test_model_manager_loads_local_models_on_init():
    with TemporaryDirectory() as tempdir:
        shutil.copytree(Fixtures.get_good_model_path(), os.path.join(tempdir, "model1"))
        shutil.copy(
            Fixtures.get_good_model_archive_path(),
            os.path.join(tempdir, "model2.zip"),
        )
        ModelManager._ModelManager__instance = None
        with temp_config(
            {"runtime": {"local_models_dir": tempdir}}, merge_strategy="merge"
        ):
            MODEL_MANAGER = ModelManager()

            assert len(MODEL_MANAGER.loaded_models) == 2
            assert "model1" in MODEL_MANAGER.loaded_models.keys()
            assert "model2.zip" in MODEL_MANAGER.loaded_models.keys()
            assert "model-does-not-exist.zip" not in MODEL_MANAGER.loaded_models.keys()


def test_load_model_error_response():
    """Test load model's model does not exist when the loader throws"""
    with pytest.raises(CaikitRuntimeException) as context:
        MODEL_MANAGER.load_model(
            model_id=random_test_id(),
            local_model_path=Fixtures().get_invalid_model_archive_path(),
            model_type="categories_esa",
            wait=True,
        )

    assert context.value.status_code == grpc.StatusCode.NOT_FOUND
    assert len(MODEL_MANAGER.loaded_models) == 0


def test_load_model_map_insertion():
    """Test if loaded model is correctly added to map storing model data"""
    model = random_test_id()
    MODEL_MANAGER.load_model(
        model_id=model,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert MODEL_MANAGER.loaded_models[model].size() > 0


def test_load_model_count():
    """Test if multiple loaded models are added to map storing model data"""
    MODEL_MANAGER.load_model(
        model_id=random_test_id(),
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    MODEL_MANAGER.load_model(
        model_id=random_test_id(),
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert len(MODEL_MANAGER.loaded_models) == 2


def test_unload_model_ok_response():
    """Test to make sure that given a loaded model ID, the model manager is able to correctly
    unload a model, giving a nonzero model size back."""
    model_id = "happy_unload_test"
    MODEL_MANAGER.load_model(
        model_id=model_id,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    model_size = MODEL_MANAGER.unload_model(model_id=model_id)
    assert model_size >= 0


def test_unload_model_count():
    """Test if unloaded models are deleted from loaded models map"""
    id_1 = "test"
    id_2 = "test2"
    # Load models from COS
    MODEL_MANAGER.load_model(
        model_id=id_1,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    MODEL_MANAGER.load_model(
        model_id=id_2,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    # Unload one of the models, and make sure it was properly removed
    MODEL_MANAGER.unload_model(model_id=id_2)
    assert len(MODEL_MANAGER.loaded_models) == 1
    assert id_1 in MODEL_MANAGER.loaded_models.keys()


def test_unload_all_models_count():
    """Test if unload all models deletes every model from loaded models map"""
    id_1 = "test"
    id_2 = "test2"
    MODEL_MANAGER.load_model(
        model_id=id_1,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    MODEL_MANAGER.load_model(
        model_id=id_2,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    MODEL_MANAGER.unload_all_models()
    assert len(MODEL_MANAGER.loaded_models) == 0


def test_unload_model_not_loaded_response():
    """Test unload model for model not loaded does NOT throw an error"""
    MODEL_MANAGER.unload_model(model_id=random_test_id())


def test_retrieve_model_returns_loaded_model():
    """Test that a loaded model can be retrieved"""
    model_id = random_test_id()
    MODEL_MANAGER.load_model(
        model_id,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),  # eventually we'd like to be determining the type from the model itself...
    )
    model = MODEL_MANAGER.retrieve_model(model_id)
    assert isinstance(model, ModuleBase)
    assert len(MODEL_MANAGER.loaded_models) == 1


def test_retrieve_model_raises_error_for_not_found_model():
    """Test that gRPC NOT_FOUND exception raised when non-existent model retrieved"""
    with pytest.raises(CaikitRuntimeException) as context:
        MODEL_MANAGER.retrieve_model("not-found")
    assert context.value.status_code == grpc.StatusCode.NOT_FOUND


def test_model_size_ok_response():
    """Test if loaded model correctly returns model size"""
    model = random_test_id()
    MODEL_MANAGER.load_model(
        model_id=model,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert MODEL_MANAGER.get_model_size(model) > 0


def test_model_size_error_not_found_response():
    """Test model size's model does not exist error response"""
    with pytest.raises(CaikitRuntimeException) as context:
        MODEL_MANAGER.get_model_size("no_exist_model")
    assert context.value.status_code == grpc.StatusCode.NOT_FOUND


def test_model_size_error_none_not_found_response():
    """Test model size's model is None error response"""
    with pytest.raises(CaikitRuntimeException) as context:
        MODEL_MANAGER.get_model_size(None)
    assert context.value.status_code == grpc.StatusCode.NOT_FOUND


def test_estimate_model_size_ok_response_on_loaded_model():
    """Test if loaded model correctly returns model size"""
    MODEL_MANAGER.load_model(
        model_id=random_test_id(),
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert (
        MODEL_MANAGER.estimate_model_size(
            "test", Fixtures.get_good_model_path(), Fixtures.get_good_model_type()
        )
        > 0
    )


def test_estimate_model_size_ok_response_on_nonloaded_model():
    """Test if a model that's not loaded, correctly returns predicted model size"""
    assert (
        MODEL_MANAGER.estimate_model_size(
            "test", Fixtures.get_good_model_path(), Fixtures.get_good_model_type()
        )
        > 0
    )


def test_estimate_model_size_by_type():
    """Test that a model's size is estimated differently based on its type"""
    config = get_config().inference_plugin.model_mesh
    assert Fixtures.get_good_model_type() in config.model_size_multipliers

    typed_model_size = MODEL_MANAGER.estimate_model_size(
        "test", Fixtures.get_good_model_path(), Fixtures.get_good_model_type()
    )
    untyped_model_size = MODEL_MANAGER.estimate_model_size(
        "test", Fixtures.get_good_model_path(), "test-not-a-model-type"
    )

    assert typed_model_size > 0
    assert untyped_model_size > 0
    assert typed_model_size != untyped_model_size


def test_estimate_model_size_error_not_found_response():
    """Test if error in predict model size on unknown model path"""
    with pytest.raises(CaikitRuntimeException) as context:
        MODEL_MANAGER.estimate_model_size(
            model_id=random_test_id(),
            local_model_path="no_exist.zip",
            model_type="categories_esa",
        )
    assert context.value.status_code == grpc.StatusCode.NOT_FOUND


def test_model_manager_replicas_with_disk_caching(good_model_path):
    """Test that multiple model manager instances can co-exist and share a set
    of models when local_models_cache is used.

    This test simulates running concurrent replicas of caikit.runtime that are
    logically independent from one another.
    """
    # NOTE: This test requires that the ModelManager class not be a singleton.
    #   To accomplish this, the singleton instance is temporarily removed.
    with TemporaryDirectory() as cache_dir:
        with non_singleton_model_managers(
            2,
            {
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": True,
                    "lazy_load_poll_period_seconds": 0,
                },
            },
            "merge",
        ) as managers:
            # Make two standalone manager instances
            manager_one, manager_two = managers
            assert manager_two is not manager_one
            assert manager_one._local_models_dir == cache_dir
            assert manager_two._local_models_dir == cache_dir

            # Trying to retrieve a model that is not loaded in either and is
            # not saved in the cache dir should yield NOT_FOUND
            model_id = random_test_id()
            with pytest.raises(CaikitRuntimeException) as context:
                manager_one.retrieve_model(model_id)
            assert context.value.status_code == grpc.StatusCode.NOT_FOUND
            with pytest.raises(CaikitRuntimeException) as context:
                manager_two.retrieve_model(model_id)
            assert context.value.status_code == grpc.StatusCode.NOT_FOUND

            # Place a saved model into the cache dir manually
            model_one_path = os.path.join(cache_dir, model_id)
            shutil.copytree(good_model_path, model_one_path)

            # Retrieve the model by ID and ensure it's available now
            model_one = manager_one.retrieve_model(model_id)
            assert isinstance(model_one, ModuleBase)
            assert (
                manager_one.loaded_models.get(model_id)
                and manager_one.loaded_models[model_id].model() is model_one
            )

            # Make sure the same model is available on the second manager
            assert manager_two.loaded_models.get(model_id) is None
            assert manager_two.retrieve_model(model_id)
            assert manager_two.loaded_models.get(model_id)

            # Remove the model from the local dir and trigger both managers to
            # sync their local model dirs to make sure it gets removed
            shutil.rmtree(model_one_path)
            manager_one.sync_local_models(wait=True)
            manager_two.sync_local_models(wait=True)
            with pytest.raises(CaikitRuntimeException) as context:
                manager_one.retrieve_model(model_id)
            with pytest.raises(CaikitRuntimeException) as context:
                manager_two.retrieve_model(model_id)


def test_model_manager_disk_caching_periodic_sync(good_model_path):
    """Make sure that when using disk caching, the manager periodically syncs
    its loaded models based on their presence in the cache
    """
    purge_period = 0.001
    with TemporaryDirectory() as cache_dir:
        with non_singleton_model_managers(
            2,
            {
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": True,
                    "lazy_load_poll_period_seconds": purge_period,
                },
            },
            "merge",
        ) as managers:
            manager_one, manager_two = managers

            # Load the model on the first manager and get it loaded into the
            # the first by fetching it
            model_id = random_test_id()
            model_cache_path = os.path.join(cache_dir, model_id)
            assert not os.path.exists(model_cache_path)
            shutil.copytree(good_model_path, model_cache_path)
            assert manager_one.retrieve_model(model_id)
            assert model_id in manager_one.loaded_models

            # Wait for the sync period and make sure it's available in the
            # second manager
            start = time.time()
            mgr_two_loaded = False
            while (time.time() - start) < (purge_period * 1000):
                time.sleep(purge_period)
                if model_id in manager_two.loaded_models:
                    mgr_two_loaded = True
                    break
            assert mgr_two_loaded

            # Remove the model from disk and wait to ensure that the model gets
            # unloaded
            shutil.rmtree(model_cache_path)
            start = time.time()
            mgr_one_unloaded = False
            mgr_two_unloaded = False
            while (time.time() - start) < (purge_period * 1000):
                if model_id not in manager_one.loaded_models:
                    mgr_one_unloaded = True
                if model_id not in manager_two.loaded_models:
                    mgr_two_unloaded = True
                if mgr_one_unloaded and mgr_two_unloaded:
                    break
            assert mgr_one_unloaded and mgr_two_unloaded


def test_load_local_model_deleted_dir():
    """Make sure losing the local_models_dir out from under a running manager
    doesn't kill the whole thing
    """
    with TemporaryDirectory() as tempdir:
        cache_dir = os.path.join(tempdir, "cache_dir")
        os.makedirs(cache_dir)
        with non_singleton_model_managers(
            1,
            {
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": True,
                    "lazy_load_poll_period_seconds": 0.001,
                },
            },
            "merge",
        ) as managers:
            manager = managers[0]

            # Make sure the timer started
            assert manager._lazy_sync_timer is not None

            # Delete the cache dir and force a sync
            shutil.rmtree(cache_dir)
            while True:
                try:
                    os.listdir(cache_dir)
                except FileNotFoundError:
                    break
            manager.sync_local_models(wait=True)

            # Make sure the timer is removed
            assert manager._lazy_sync_timer is None


def test_load_local_model_deleted_dir():
    """Make sure bad models in local_models_dir at boot don't cause exceptions"""
    with TemporaryDirectory() as cache_dir:
        model_id = random_test_id()
        model_cache_path = os.path.join(cache_dir, model_id)
        os.makedirs(model_cache_path)
        with non_singleton_model_managers(
            1,
            {
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": False,
                },
            },
            "merge",
        ) as managers:
            manager = managers[0]
            assert not manager.loaded_models


def test_lazy_load_ephemeral_model():
    """Make sure an ephemeral model (not on disk) can be lazy loaded if the
    right finder configuration is present to load it without hitting disk.
    """
    with TemporaryDirectory() as cache_dir:
        with non_singleton_model_managers(
            1,
            {
                "model_management": {"finders": {"default": {"type": TestFinder.name}}},
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": True,
                },
            },
            "merge",
        ) as managers:
            manager = managers[0]
            model_id = random_test_id()
            model = manager.retrieve_model(model_id)
            assert model
            assert cache_dir not in manager.loaded_models[model_id].path()

            # Make sure the model does not get unloaded on sync
            manager._local_models_dir_sync(wait=True)
            assert model_id in manager.loaded_models


# ****************************** Unit Tests ****************************** #
# These tests patch in mocks for the manager's dependencies, to test its code in isolation


def test_load_model():
    """Test to make sure that given valid input, the model manager gives a happy response
    when we tried to load in a model (model size > 0 or 0 if the model size will be computed
    at a later time)."""
    mock_loader = MagicMock()
    mock_sizer = MagicMock()
    model_id = random_test_id()
    expected_model_size = 1234

    with patch.object(MODEL_MANAGER, "model_loader", mock_loader):
        with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
            loaded_model = LoadedModel()
            loaded_model._model = "something"
            mock_loader.load_model.return_value = loaded_model
            mock_sizer.get_model_size.return_value = expected_model_size

            model_size = MODEL_MANAGER.load_model(
                model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE
            )
            assert expected_model_size == model_size
            mock_loader.load_model.assert_called_once()
            call_args = mock_loader.load_model.call_args
            assert call_args.args == (
                model_id,
                ANY_MODEL_PATH,
                ANY_MODEL_TYPE,
            )
            assert call_args.kwargs["aborter"] is None
            assert "fail_callback" in call_args.kwargs
            mock_sizer.get_model_size.assert_called_once_with(
                model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE
            )


def test_load_model_throws_if_the_model_loader_throws():
    """Test load model's model does not exist when the loader throws"""
    mock_loader = MagicMock()
    with patch.object(MODEL_MANAGER, "model_loader", mock_loader):
        mock_loader.load_model.side_effect = CaikitRuntimeException(
            grpc.StatusCode.NOT_FOUND, "test any not found exception"
        )

        with pytest.raises(CaikitRuntimeException) as context:
            MODEL_MANAGER.load_model(random_test_id(), ANY_MODEL_PATH, ANY_MODEL_TYPE)

        assert context.value.status_code == grpc.StatusCode.NOT_FOUND
        assert len(MODEL_MANAGER.loaded_models) == 0


def test_retrieve_model_returns_the_module_from_the_model_loader():
    """Test that a loaded model can be retrieved"""
    model_id = random_test_id()
    expected_module = "this is definitely a stub module"
    mock_sizer = MagicMock()
    mock_loader = MagicMock()

    with patch.object(MODEL_MANAGER, "model_loader", mock_loader):
        with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
            mock_sizer.get_model_size.return_value = 1
            model_future = Future()
            model_future.result = lambda *_, **__: expected_module
            model_future_factory = lambda: model_future
            mock_loader.load_model.return_value = (
                LoadedModel.Builder()
                .model_future_factory(model_future_factory)
                .id("foo")
                .type("bar")
                .build()
            )
            MODEL_MANAGER.load_model(model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE)

            model = MODEL_MANAGER.retrieve_model(model_id)
            assert expected_module == model


def test_unload_partially_loaded():
    """Make sure that when unloading a model that is still loading, the load
    completes correctly
    """
    # Set up a "slow load" that we can use to ensure that the loading has
    # completed successfully
    slow_loader = SlowLoader()
    pool = ThreadPoolExecutor()
    model_future_factory = partial(pool.submit, slow_loader.load)

    model_id = random_test_id()
    mock_sizer = MagicMock()
    mock_loader = MagicMock()
    with patch.object(MODEL_MANAGER, "model_loader", mock_loader):
        with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
            mock_sizer.get_model_size.return_value = 1
            mock_loader.load_model.return_value = (
                LoadedModel.Builder()
                .model_future_factory(model_future_factory)
                .id("foo")
                .type("bar")
                .build()
            )
            MODEL_MANAGER.load_model(
                model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE, wait=False
            )

            # Start the unload assertion that will block on loading
            unload_future = pool.submit(MODEL_MANAGER.unload_model, model_id)

            # Unblock the load
            assert not slow_loader.done_loading()
            slow_loader.unblock_load()

            # Make sure unload completes and the model finished loading
            unload_future.result()
            assert slow_loader.done_loading()


def test_unload_unexpected_error_loaded():
    """Make sure that when unloading a model that is still loading, errors when
    waiting for the load to complete are handled
    """
    # Set up a "slow load" that we can use to ensure that the loading has
    # completed successfully
    slow_loader = SlowLoader(RuntimeError("yikes"))
    pool = ThreadPoolExecutor()
    model_future_factory = partial(pool.submit, slow_loader.load)

    model_id = random_test_id()
    mock_sizer = MagicMock()
    mock_loader = MagicMock()
    with patch.object(MODEL_MANAGER, "model_loader", mock_loader):
        with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
            mock_sizer.get_model_size.return_value = 1
            mock_loader.load_model.return_value = (
                LoadedModel.Builder()
                .model_future_factory(model_future_factory)
                .id("foo")
                .type("bar")
                .build()
            )
            MODEL_MANAGER.load_model(
                model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE, wait=False
            )

            # Start the unload assertion that will block on loading
            unload_future = pool.submit(MODEL_MANAGER.unload_model, model_id)

            # Unblock the load
            assert not slow_loader.done_loading()
            slow_loader.unblock_load()

            # Make sure unload completes and the model finished loading
            with pytest.raises(CaikitRuntimeException) as context:
                unload_future.result()
            assert context.value.status_code == grpc.StatusCode.INTERNAL
            assert slow_loader.done_loading()


def test_reload_partially_loaded():
    """Make sure that attempting to reload a model that has already started to
    load simply returns the existing LoadedModel
    """
    # Set up a "slow load" that we can use to ensure that the loading has
    # completed successfully
    slow_loader = SlowLoader()
    pool = ThreadPoolExecutor()
    model_future_factory = partial(pool.submit, slow_loader.load)

    model_id = random_test_id()
    mock_sizer = MagicMock()
    mock_loader = MagicMock()
    special_model_size = 123321
    with patch.object(MODEL_MANAGER, "model_loader", mock_loader):
        with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
            mock_sizer.get_model_size.return_value = special_model_size
            loaded_model = (
                LoadedModel.Builder()
                .model_future_factory(model_future_factory)
                .id("foo")
                .type("bar")
                .build()
            )
            mock_loader.load_model.return_value = loaded_model
            model_size = MODEL_MANAGER.load_model(
                model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE, wait=False
            )
            assert model_size == special_model_size
            assert (
                MODEL_MANAGER.load_model(
                    model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE, wait=False
                )
                == special_model_size
            )

            # Unblock the load
            assert not slow_loader.done_loading()
            slow_loader.unblock_load()
            loaded_model.wait()
            assert slow_loader.done_loading()


def test_get_model_size_returns_size_from_model_sizer():
    """Test that loading a model stores the size from the ModelSizer"""
    mock_loader = MagicMock()
    mock_sizer = MagicMock()
    expected_model_size = 1234
    model_id = random_test_id()

    with patch.object(MODEL_MANAGER, "model_loader", mock_loader):
        with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
            loaded_model = LoadedModel()
            loaded_model._model = "something"
            mock_loader.load_model.return_value = loaded_model
            mock_sizer.get_model_size.return_value = expected_model_size

            MODEL_MANAGER.load_model(model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE)

            model_size = MODEL_MANAGER.get_model_size(model_id)
            assert expected_model_size == model_size


def test_estimate_model_size_returns_size_from_model_sizer():
    """Test that estimating a model size uses the ModelSizer"""
    mock_sizer = MagicMock()
    expected_model_size = 5678
    model_id = random_test_id()

    with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
        mock_sizer.get_model_size.return_value = expected_model_size
        model_size = MODEL_MANAGER.estimate_model_size(
            model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE
        )
        assert expected_model_size == model_size


def test_estimate_model_size_throws_if_model_sizer_throws():
    """Test that estimating a model size uses the ModelSizer"""
    mock_sizer = MagicMock()
    model_id = random_test_id()

    with patch.object(MODEL_MANAGER, "model_sizer", mock_sizer):
        mock_sizer.get_model_size.side_effect = CaikitRuntimeException(
            grpc.StatusCode.UNAVAILABLE, "test-any-exception"
        )
        with pytest.raises(CaikitRuntimeException) as context:
            MODEL_MANAGER.estimate_model_size(model_id, ANY_MODEL_PATH, ANY_MODEL_TYPE)
        assert context.value.status_code == grpc.StatusCode.UNAVAILABLE


def test_periodic_sync_handles_errors():
    """Test that any exception raised during syncing local models is handled
    without terminating the polling loop
    """

    class SecretException(Exception):
        pass

    with TemporaryDirectory() as cache_dir:
        with non_singleton_model_managers(
            1,
            {
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": True,
                },
            },
            "merge",
        ) as managers:
            manager = managers[0]
            with patch.object(manager, "_local_models_dir_sync") as mock_sync:
                mock_sync.side_effect = SecretException()
                assert manager._lazy_sync_timer is not None
                manager.sync_local_models(True)
                mock_sync.assert_called_once()
                assert manager._lazy_sync_timer is not None


def test_periodic_sync_handles_temporary_errors():
    """Test that models loaded with the periodic sync can retry if the initial
    load operation fails
    """
    with TemporaryDirectory() as cache_dir:
        with non_singleton_model_managers(
            1,
            {
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": True,
                    "lazy_load_retries": 1,
                },
            },
            "merge",
        ) as managers:
            manager = managers[0]
            flakey_loader = TempFailWrapper(
                manager.model_loader._load_module,
                num_failures=1,
                exc=CaikitRuntimeException(grpc.StatusCode.INTERNAL, "Dang"),
            )
            with patch.object(
                manager.model_loader,
                "_load_module",
                flakey_loader,
            ):
                assert manager._lazy_sync_timer is not None
                model_path = Fixtures.get_good_model_path()
                model_name = os.path.basename(model_path)
                shutil.copytree(model_path, os.path.join(cache_dir, model_name))
                manager.sync_local_models(True)
                assert manager._lazy_sync_timer is not None
                model = manager.retrieve_model(model_name)
                assert model


def test_lazy_load_handles_temporary_errors():
    """Test that a lazy load without a periodic sync correctly retries failed
    loads
    """
    with TemporaryDirectory() as cache_dir:
        with non_singleton_model_managers(
            1,
            {
                "runtime": {
                    "local_models_dir": cache_dir,
                    "lazy_load_local_models": True,
                    "lazy_load_poll_period_seconds": 0,
                    "lazy_load_retries": 1,
                },
            },
            "merge",
        ) as managers:
            manager = managers[0]
            flakey_loader = TempFailWrapper(
                manager.model_loader._load_module,
                num_failures=1,
                exc=CaikitRuntimeException(grpc.StatusCode.INTERNAL, "Dang"),
            )
            with patch.object(
                manager.model_loader,
                "_load_module",
                flakey_loader,
            ):
                assert manager._lazy_sync_timer is None
                model_path = Fixtures.get_good_model_path()
                model_name = os.path.basename(model_path)
                shutil.copytree(model_path, os.path.join(cache_dir, model_name))
                assert manager._lazy_sync_timer is None
                model = manager.retrieve_model(model_name)
                assert model



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A sample module for sample things!
"""
# Local
from ...data_model.sample import FileDataType, FileTask
from caikit.core.modules import ModuleLoader
import caikit.core


@caikit.core.module(
    "750cad4d-c3b8-4327-b52e-e772f0d6f311", "BoundingBoxModule", "0.0.1", FileTask
)
class BoundingBoxModule(caikit.core.ModuleBase):
    def run(
        self,
        unprocessed: FileDataType,
    ) -> FileDataType:
        filename = f"processed_{unprocessed.filename}"
        data = b"bounding|" + unprocessed.data + b"|box"
        return FileDataType(filename, data)

    @classmethod
    def load(cls, model_path, **kwargs):
        loader = ModuleLoader(model_path)
        config = loader.config
        return cls()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A generic module to help import dynamic modules
"""

# Standard
from types import ModuleType
from typing import Any, List
import importlib
import re
import sys

# Third Party
from grpc import StatusCode

# First Party
import aconfig
import alog

# Local
from caikit import get_config
from caikit.core import MODEL_MANAGER
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
import caikit.core

log = alog.use_channel("COM-LIB-INIT")


class UnifiedDataModel:
    """This class acts as an aggregator between the data models of multiple
    Caikit libraries.
    """

    def __init__(self):
        """Initialize the internal set of libraries"""
        self._libraries = {}

    def add_library(self, lib_name: str, lib: ModuleType):
        """Add the given library to the unified data model"""
        if lib_name in self._libraries:
            raise ValueError(f"Double registration of {lib_name}")
        self._libraries[lib_name] = getattr(lib, "data_model", None)

    def __getattr__(self, name: str) -> Any:
        """Fetch an attribute by name aliasing into child libraries"""
        if name in self._libraries:
            return self._libraries[name]
        candidates = {
            lib_name: getattr(lib, name)
            for lib_name, lib in self._libraries.items()
            if hasattr(lib, name)
        }
        if len(candidates) == 1:
            return list(candidates.values())[0]
        if len(candidates) > 1:
            raise AttributeError(
                f"Multiple library implementations of {name} found: {list(candidates.keys())}"
            )
        return super().__getattr__(name)


def get_data_model(config: aconfig.Config = None) -> UnifiedDataModel:
    """
    Get the data model from the Caikit library of interest. This is accomplished
    via dynamic import on the caikit_library's environment variable.

    NOTE: This function also has the side-effect of importing each of the
        caikit_library libraries for the first time, causing their modules to
        be registered with the caikit.core module registry. It is a critical
        step in initializing the set of modules that can be loaded by this
        running server instance.

    Args:
        config(aconfig.Config): caikit configuration

    Returns:
        (module): Handle to the module after dynamic wild import
    """
    config = config or get_config()
    lib_names = clean_lib_names(config.runtime.library)

    # Add all caikit.interfaces.X modules
    lib_names.extend(
        [
            lib_name
            for lib_name in sys.modules
            if lib_name.startswith("caikit.interfaces.") and lib_name.count(".") == 2
        ]
    )

    cdm = UnifiedDataModel()
    for lib_name in lib_names:
        cdm = _get_cdm_from_lib(lib_name, cdm)

    # Check module registry to get base modules
    # NOTE: Following is done after previous for-loop, since we need to import
    # all the libraries which will register all the modules and that happens
    # in get_dynamic_module above
    base_lib_names = set()
    for module_class in caikit.core.registries.module_registry().values():
        lib_name = module_class.__module__.partition(".")[0]
        if lib_name not in lib_names:
            # This module is from a library not mentioned
            # in lib_names. Consider this as a base library, like caikit.core
            base_lib_names.add(lib_name)

    # Get data model from lib_names
    for lib_name in base_lib_names:
        cdm = _get_cdm_from_lib(lib_name, cdm)

    # Ensure that all model management components have been initialized
    # TODO: This function has a _ton_ of side effects! We need to split it up
    #   and isolate these side effects in appropriately named functions.
    #   Specifically, this function is not only responsible for creating the
    #   data model, but it performs the dynamic import of the domain library and
    #   initializes the model management components.
    MODEL_MANAGER.initialize_components()

    return cdm


def _get_cdm_from_lib(lib_name: str, cdm: UnifiedDataModel):
    """Function to get caikit core CDM from library name

    Args:
        lib_name (str): Caikit core library name
        cdm (UnifiedDataModel): Caikit core CDM
    Returns:
        cdm: UnifiedDataModel
    """
    caikit_library = get_dynamic_module(lib_name)

    if caikit_library is None:
        message = "Unable to load data model from library: %s" % (lib_name)
        log.error("<RUN22291311E>", message)
        raise ValueError(message)
    cdm.add_library(lib_name, caikit_library)
    return cdm


def get_dynamic_module(module_name: str, module_dir: str = None) -> ModuleType:
    """
    Get the dynamic module of interest.

    Args:
        module_name(str): Name of the module to be dynamically imported
        (Optional) module_dir(str): Name of the directory from where the module is
                to be dynamically imported

    Returns:
        (module): Handle to the module after dynamic import
    """
    if module_dir:
        module_path = module_dir + "." + module_name
    else:
        module_path = module_name
    log.info("<RUN11997772I>", "Loading service module: %s", module_path)
    # Try to find the spec for the module that we're interested in.
    spec = importlib.util.find_spec(module_path)
    if not spec:
        message = "Unable to find spec for module: %s" % (module_path)
        # TODO: figure out the better way of doing this
        # https://github.com/caikit/caikit/pull/85#discussion_r1182890609
        log.warning("<RUN11991313W>", message)
        raise CaikitRuntimeException(StatusCode.NOT_FOUND, message)
    # Found spec - import the library
    if module_dir:
        return importlib.import_module(module_path)

    return importlib.import_module(module_path, "*")


def clean_lib_names(caikit_library: str) -> List[str]:
    def clean(lib):
        # Regex explanation:
        # - Capturing group ([a-zA-Z_-]+)
        #    To match 1 or more of alphabets (small or uppercase) and symbols (-) and (_)
        # - Match special character `[` (0 or 1) times
        # - Match any number of characters present in the list [a-zA-Z_\.,]
        #   NOTE: . is special character thus prefixed by \
        # - Match special character `]` (0 or 1) times
        # - Match on of following symbols [<>=\d\.\-a-z] (\d is for digits)
        regex = r"([a-zA-Z_-]+)\[?[a-zA-Z_\.,]*\]?[<>=\d\.\-a-z]*"
        lib_name = re.search(regex, lib).group(1)
        cleaned = lib_name.replace("-", "_")  # replace hyphens with underscores
        return cleaned

    lib_names = caikit_library.split()
    return [clean(lib) for lib in lib_names]



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The ReversibleHasher provides a simple obfuscation tool for munging strings
in a reversible way. It is useful when needing a repeatable way to obfuscate a
string that may come from a configuration detail that should not be
transparently visible to a user, but needs to be trivially reversed, possibly
with entirely new in-memory state.

WARNING: This utility is intentionally not a cryptographically secure hash! It
    is explicitly designed to be reversible.
"""

# Standard
import random
import string

# Create a statically seeded instance of Random to perform shuffling in a
# consistent way
_RAND_INST = random.Random()
_RAND_INST.seed(42)
_CHARSET = string.ascii_letters + string.digits + "-_"
_SHUFFLED = list(_CHARSET)
_RAND_INST.shuffle(_SHUFFLED)

# Create the static forward/backward maps for the character set
FORWARD_MAP = dict(zip(_CHARSET, _SHUFFLED))
BACKWARD_MAP = {v: k for k, v in FORWARD_MAP.items()}


class ReversibleHasher:
    __doc__ = __doc__

    @classmethod
    def hash(cls, val: str):
        return "".join([FORWARD_MAP.get(ch, ch) for ch in val])

    @classmethod
    def reverse_hash(cls, hash_val: str):
        return "".join([BACKWARD_MAP.get(ch, ch) for ch in hash_val])



"""
CAIKit is an AI toolkit that enables AI users to consume stable task-specific
model APIs while enabling AI developers build algorithms and models in a
modular/composable framework.
"""
# Local
from . import core, interfaces

# Expose configuration fn and getter at the top level
from .config import configure, get_config

# Expose model management at the top level
from .core import extract, get_model_future, load, module, resolve_and_load, train



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Local
from ...data_model import SampleInputType, SampleOutputType, SampleTask
from ...modules.sample_task.sample_implementation import SampleModule
from caikit.core.modules import ModuleLoader, ModuleSaver
import caikit


@caikit.module(
    "A34E68FA-E5E6-41BD-BAAE-77A880EB6877", "CompositeModule", "0.0.1", SampleTask
)
class CompositeModule(caikit.core.ModuleBase):
    def __init__(self, sample_block: SampleModule = SampleModule()):
        super().__init__()
        self.block = sample_block

    def run(self, sample_input: SampleInputType) -> SampleOutputType:
        """Runs the inner block"""
        return self.block.run(sample_input)

    @classmethod
    def load(cls, model_path: str):
        loader = ModuleLoader(model_path)
        return CompositeModule(loader.load_module("dummy_model"))

    def save(self, model_path):
        saver = ModuleSaver(
            module=self,
            model_path=model_path,
        )
        with saver:
            saver.save_module(self.block, "dummy_model")

    @classmethod
    def train(cls, sample_block: SampleModule) -> "CompositeModule":
        return cls(sample_block)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""DataStream converter that converts a given Datastream to either a DataStream of lists or
dictionaries.
These type conversions are encapsulated here right now for consistency in the workflows served in
production cloud offerings, but we could see moving these to be directly on the DataStream class.
"""

# Standard
from typing import Callable, Dict, List

# First Party
import alog

# Local
from ...exceptions import error_handler
from .. import DataStream

log = alog.use_channel("DATSTRMCNVRTR")
error = error_handler.get(log)


class DataStreamConverter:
    """Converts DataStreams to a target type

    This uses a target type and a list of expected keys to convert each data item in the stream.

    For example, for an input stream that looks like:
        [
            { a: 1, b: 2, c: 3 }
            { a: 4, b: 5, c: 6 }
            { a: 7, b: 8, c: 9 }
        ]
    If target_type is list with key_list of ['a','c'], the result will be:
        [
            [1, 3]
            [4, 6]
            [7, 9]
        ]

    Or, for an input stream that looks like:
        [
            [1, 3]
            [4, 6]
            [7, 9]
        ]
    If target_type is dict with key_list of ['foo', 'bar'], the result will be:
        [
            { foo: 1, bar: 3 }
            { foo: 4, bar: 6 }
            { foo: 7, bar: 9 }
        ]

    >>> list_of_dicts = [{"foo": 1, "bar": 2}, {"foo": 3, "bar": 4}]
    >>> dict_stream = DataStream.from_iterable(list_of_dicts)
    >>> converter = DataStreamConverter(target_type=list, key_list=['foo', 'bar'])
    >>> list_stream = converter.convert(dict_stream)
    """

    def __init__(self, target_type: type, key_list: List[str]):
        """Initialize DataStreamConverter

        Args:
            target_type (type): Target DataStream type, either dict or list
            key_list (List(str)): List of keys that determines how data will be
                formatted into a DataStream
        """
        error.type_check("<COR56775827E>", type, target_type=target_type)
        error.type_check_all("<COR16523028E>", str, key_list=key_list)
        # Tuple support can be added here in the future if anybody needs it
        self._conversion_function_map: Dict[type, Callable] = {
            dict: self._convert_stream_to_dicts,
            list: self._convert_stream_to_lists,
        }

        # only support streams of dict or list for now
        if target_type not in self._conversion_function_map:
            error(
                "<COR77775827E>",
                ValueError(
                    "Conversion of stream data items to type {} not supported".format(
                        target_type
                    )
                ),
            )

        self._target_type: type = target_type
        self._key_list: List[str] = key_list
        error.value_check(
            "<COR98237539E>", self._key_list, "`key_list` should be nonempty"
        )

    def convert(self, stream: DataStream) -> DataStream:
        """Attempt to convert a given datastream to a datastream of the target type
        See classdoc for examples

        Args:
            stream (DataStream): stream intended to be converted

        Returns:
            Converted datastream based on the target type
        """
        # check the target type
        return self._conversion_function_map[self._target_type](stream)

    def _convert_stream_to_dicts(self, stream: DataStream) -> DataStream:
        """Attempt to convert a stream to dictionaries

        Args:
            stream (DataStream): Stream to convert to a stream of dictionaries
        Returns:
            A stream which will lazily convert data items to dictionaries
        """

        def convert(data_item):
            if isinstance(data_item, dict):
                log.debug(
                    "Trying to convert stream to dict, but data item was already a dict"
                )
                return data_item

            return {key: data_item[index] for index, key in enumerate(self._key_list)}

        return stream.map(convert)

    def _convert_stream_to_lists(self, stream: DataStream) -> DataStream:
        """Attempt to convert a stream to lists

        Args:
            stream (DataStream): Stream to convert to a stream of lists
        Returns:
            A stream which will lazily convert data items to lists
        """

        def convert(data_item):
            if isinstance(data_item, list):
                log.debug(
                    "Trying to convert stream to list, but data item was already a list"
                )
                return data_item

            if isinstance(data_item, tuple):
                log.debug("Converting tuple to list")
                return list(data_item)

            return [data_item[key] for key in self._key_list]

        return stream.map(convert)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the model Batcher
"""

# Standard
import os
import queue
import threading
import time

# Third Party
import pytest

# First Party
import alog

# Local
from caikit.runtime.model_management.batcher import Batcher
from sample_lib.data_model import SampleOutputType
import caikit.core

## Helpers #####################################################################

log = alog.use_channel("TEST")

DUMMY_MODEL = os.path.realpath(
    os.path.join(
        os.path.dirname(__file__),
        "..",
        "fixtures",
        "models",
        "foo_archive.zip",
    )
)


@caikit.core.module(
    "7464f684-58e3-4e99-9a58-1c5bc085472b", "slow sample module", "0.0.1"
)
class SlowSampleModule(caikit.core.ModuleBase):
    """This module is just a wrapper around another module that will inject a
    sleep
    """

    def __init__(self, model, sleep_time_s=0.1):
        self._model = model
        self._sleep_time_s = sleep_time_s
        self.batches = []
        self.batches_lock = threading.Lock()

    def run(self, *args, **kwargs):
        log.debug("Starting slow module")
        time.sleep(self._sleep_time_s)
        return self._model.run(*args, **kwargs)

    def run_batch(self, *args, **kwargs):
        with self.batches_lock:
            self.batches.append((args, kwargs))
        return self._model.run_batch(*args, **kwargs)


@caikit.core.module(
    "19b126aa-b55b-4349-94f1-d676f3e12c9b",
    "Really silly bunch of bobs!",
    "0.0.1",
)
class StubModule(caikit.core.ModuleBase):
    # NOTE: The initial implementation had "num_bobs" which expected an int.
    #   This led to the discovery of the fact that lists of ints are not
    #   considered "expandable iterables" in the default run_batch impl.
    def __init__(self):
        super().__init__()
        self.reqs = []

    def run(self, bobs_name, last_name="Bit"):
        self.reqs.append({"bobs_name": bobs_name, "last_name": last_name})
        return SampleOutputType(greeting=f"{bobs_name} {last_name}")


class ModelRunThread(threading.Thread):
    def __init__(self, model, run_num, request_kwargs=None):
        super().__init__()
        self.model = model
        self.run_num = run_num
        self.response = None
        self.request_kwargs = (
            request_kwargs
            if request_kwargs is not None
            else {
                "producer_id": caikit.core.data_model.ProducerId(
                    str(self.run_num),
                    "1.2.3",
                )
            }
        )

    def run(self):
        log.debug4("Running model with num %d", self.run_num)
        try:
            self.response = self.model.run(**self.request_kwargs)
            log.debug4("Finished run num %d: %s", self.run_num, self.response)
        except Exception as err:
            self.response = err


class ModelRunWrapperThread(threading.Thread):
    """Another layer of thread wrapping that will actually do the run in a sub-
    thread so that the parent thread can know when run has started without
    waiting for it to complete.
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.started_event = threading.Event()
        self.run_thread = ModelRunThread(*args, **kwargs)

    @property
    def response(self):
        return self.run_thread.response

    def run(self):
        log.debug2("Starting sub-thread")
        self.run_thread.start()
        log.debug2("Marking run as started")
        self.started_event.set()
        log.debug2("Waiting for sub-thread to complete")
        self.run_thread.join()


class BlockedSimpleQueue(queue.SimpleQueue):
    """This child class of SimpleQueue allows the queue to block all put calls
    on a manually set event so that we can simulate race conditions.
    """

    def __init__(self, *args, **kwargs):
        self._event = threading.Event()
        super().__init__(*args, **kwargs)

    def proceed(self):
        self._event.set()

    def put(self, *args, **kwargs):
        self._event.wait()
        return super().put(*args, **kwargs)


## Tests #######################################################################


@pytest.skip(
    "Some tests here tend to deadlock! Will need debugging but skipping for now",
    allow_module_level=True,
)
def test_single_request():
    """Test that a single request to a batched model acts the same as a request
    to the original model
    """
    model = caikit.core.load(DUMMY_MODEL)
    wrapped_model = Batcher("test-model", model, 10)
    prod_id = caikit.core.data_model.ProducerId("test", "1.2.3")
    standard_res = model.run(producer_id=prod_id)
    wrapped_res = wrapped_model.run(producer_id=prod_id)
    assert standard_res.to_dict() == wrapped_res.to_dict()


def test_multi_request_batch():
    """Make sure that batches of size > 1 can be run"""
    model = SlowSampleModule(caikit.core.load(DUMMY_MODEL))
    batch_size = 10
    wrapped_model = Batcher("test-model", model, batch_size)
    threads = {str(i): ModelRunThread(wrapped_model, i) for i in range(2 * batch_size)}

    for thread in threads.values():
        thread.start()

    for thread in threads.values():
        thread.join()

    # Make sure the right requests got the right responses
    for exp_pid_name, thread in threads.items():
        assert exp_pid_name == thread.response.producer_id.name

    # Make sure that non-trivial batches were run
    assert model.batches
    assert any(len(batch) > 1 for batch in model.batches)


def test_stop_in_flight_batch():
    """Make sure that setting the stop event will preempt running a batch"""
    model_delay = 0.001
    model = SlowSampleModule(caikit.core.load(DUMMY_MODEL), model_delay)
    batch_size = 10
    wrapped_model = Batcher("test-model", model, batch_size, model_delay * 2)
    threads = {str(i): ModelRunThread(wrapped_model, i) for i in range(2 * batch_size)}

    for thread in threads.values():
        thread.start()

    # Stop the
    wrapped_model.stop()

    # Make sure all of the threads finish
    for thread in threads.values():
        thread.join()

    # Make sure all requests that didn't get run got an appropriate error. This
    # is somewhat nondeterministic based on the threading environment, so we
    # can't guarantee that _any_ requests failed since they _might_ have all
    # finished before stop was able to interrupt them.
    successful_reqs = [
        prod_id.name for batch in model.batches for prod_id in batch[1]["producer_id"]
    ]
    for exp_pid_name, thread in threads.items():
        if exp_pid_name in successful_reqs:
            assert not isinstance(thread.response, Exception)
        else:
            assert isinstance(thread.response, RuntimeError)


def test_destructor():
    """Make sure that when the destructor for a model is called, everything
    shuts down cleanly, including when the model is already stopped or never
    started.
    """
    model = caikit.core.load(DUMMY_MODEL)

    # Delete before starting the thread
    wrapped_model = Batcher("test-model", model, 10)
    del wrapped_model

    # Delete after a call
    wrapped_model = Batcher("test-model", model, 10)
    prod_id = caikit.core.data_model.ProducerId("test", "1.2.3")
    wrapped_model.run(producer_id=prod_id)
    del wrapped_model

    # Delete after a call and stopping
    wrapped_model = Batcher("test-model", model, 10)
    wrapped_model.run(producer_id=prod_id)
    wrapped_model.stop()
    del wrapped_model


def test_no_restart():
    """Make sure that a model cannot be restarted once it has stopped"""
    model = caikit.core.load(DUMMY_MODEL)

    # Delete after a call and stopping
    wrapped_model = Batcher("test-model", model, 10)
    prod_id = caikit.core.data_model.ProducerId("test", "1.2.3")
    wrapped_model.run(producer_id=prod_id)
    wrapped_model.stop()
    with pytest.raises(RuntimeError):
        wrapped_model.run(producer_id=prod_id)


def test_different_kwargs():
    """Make sure that calls with ragged kwargs are batched correctly"""

    model = StubModule()
    batch_size = 5
    wrapped_model = Batcher("stub-model", model, batch_size)
    threads = {
        i: (
            ModelRunThread(wrapped_model, i, {"bobs_name": "Bob"})
            if i % 2
            else ModelRunThread(
                wrapped_model, i, {"bobs_name": "Jim", "last_name": "Bo"}
            )
        )
        for i in range(2 * batch_size)
    }

    for thread in threads.values():
        thread.start()

    for thread in threads.values():
        thread.join()

    # Make sure that the batches didn't contain args for "last_name" for odd
    # requests
    for req in model.reqs:
        if req["bobs_name"] == "Bob":
            assert req["last_name"] == "Bit"  # The default value
        else:
            assert req["last_name"] == "Bo"


def test_invalid_req_after_valid_req():
    """Make sure that in a batch, if a request is processed with a missing
    required keyword arg after a request that has it, the valid request is
    processed and the invalid request is rejected.
    """

    model = SlowSampleModule(StubModule())
    batch_size = 5
    wrapped_model = Batcher("stub-model", model, batch_size)
    threads = {
        i: (
            ModelRunThread(wrapped_model, i, {"bobs_name": "Bob", "last_name": str(i)})
            if i < 2
            else ModelRunThread(wrapped_model, i, {})
        )
        for i in range(2 * batch_size)
    }

    for thread in threads.values():
        thread.start()

    for thread in threads.values():
        thread.join()

    # Make sure that all requests after the first two had errors, but the first
    # two did not
    for i, thread in threads.items():
        if i < 2:
            assert not isinstance(thread.response, Exception)
            assert thread.response.bobs[0].split()[-1] == str(i)
        else:
            assert isinstance(thread.response, RuntimeError)


def test_valid_req_after_invalid_req():
    """Make sure that in a batch, if a request is processed that adds new
    keyword args which previous requests didn't have that also don't have
    defaults, the previous invalid requests are rejected.
    """

    model = SlowSampleModule(StubModule())
    batch_size = 5
    wrapped_model = Batcher("stub-model", model, batch_size)
    threads = {
        i: (
            ModelRunThread(wrapped_model, i, {"bobs_name": "Bob", "last_name": str(i)})
            if i >= 2
            else ModelRunThread(wrapped_model, i, {})
        )
        for i in range(2 * batch_size)
    }

    for thread in threads.values():
        thread.start()

    for thread in threads.values():
        thread.join()

    # Make sure that all requests after the first two had errors, but the first
    # two did not
    for i, thread in threads.items():
        if i >= 2:
            assert not isinstance(thread.response, Exception)
            assert thread.response.bobs[0].split()[-1] == str(i)
        else:
            assert isinstance(thread.response, RuntimeError)


def test_batch_collect_delay():
    """Make sure that with the batch delay, multiple sequential calls end up in
    the same batchg even when not presented concurrently
    """
    model = SlowSampleModule(StubModule())
    batch_size = 5
    batch_collect_delay = 0.01
    wrapped_model = Batcher("stub-model", model, batch_size, batch_collect_delay)

    # Start one thread, wait for less than the collect delay, then start another
    th1 = ModelRunThread(wrapped_model, 0, {"bobs_name": "Bob"})
    th1.start()
    log.debug("Sleeping after th1.start()")
    time.sleep(batch_collect_delay / 10)
    log.debug("Done sleeping")
    th2 = ModelRunThread(wrapped_model, 1, {"bobs_name": "Jim"})
    th2.start()
    log.debug("th2 started")

    # Wait for both to complete
    th1.join()
    th2.join()

    # Make sure only one batch was sent
    assert len(model.batches) == 1


def test_orphaned_events_race():
    """There's a subtle race condition that can heppen when handling a request
    where the event is created, stop is called and completes, _then_ the request
    gets added to the queue and will never be pulled out by stop() if it has
    finished. To avoid orphaned requests in the queue, we ensure that the event
    is somewhere that stop() can find it by putting it in a member dict on
    creation. This tests ensures that the race is mitigated.
    """
    wrapped_model = Batcher("stub-model", StubModule(), 10)

    # Patch the batcher's run function to wait on an intentional condition after
    # creating the event
    blocked_queue = BlockedSimpleQueue()
    wrapped_model._input_q = blocked_queue

    # Create a thread that will add a request. When run, this will create the
    # event and start to add to the queue, but it won't proceed until the queue
    # is unblocked.
    log.debug("Starting request thread")
    th = ModelRunWrapperThread(wrapped_model, 0, {"bobs_name": "Bob"})
    th.start()

    # Wait until the "run" call has started
    log.debug("Waiting for thread to start")
    th.started_event.wait()

    # Call stop to attempt to shut down all requests. The event is still in
    # limbo since it hasn't finished being added to the queue.
    log.debug("Stopping wrapped model")
    wrapped_model.stop()

    # Let the queue's put proceed
    log.debug("Unblocking queue.put")
    blocked_queue.proceed()

    # Wait for the request to complete. If the race exists, this will deadlock!
    log.debug("Joining request")
    th.join()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import os

# Local
from caikit.core.data_model.streams import multipart_decoder


def test_simple_multipart_file(sample_multipart_json):
    assert multipart_decoder.is_multipart_file(sample_multipart_json)
    parts = multipart_decoder.stream_multipart_file(sample_multipart_json)

    parts = list(parts)
    assert len(parts) == 1

    byte_content = parts[0].fp.read()

    with open(sample_multipart_json, "rb") as fp:
        full_multipart_byte_content = fp.read()

    assert byte_content in full_multipart_byte_content


def test_multipart_file_with_content_header(sample_multipart_json_with_content_header):
    assert multipart_decoder.is_multipart_file(
        sample_multipart_json_with_content_header
    )
    parts = multipart_decoder.stream_multipart_file(
        sample_multipart_json_with_content_header
    )

    parts = list(parts)
    assert len(parts) == 1

    byte_content = parts[0].fp.read()

    with open(sample_multipart_json_with_content_header, "rb") as fp:
        full_multipart_byte_content = fp.read()

    assert byte_content in full_multipart_byte_content


def test_many_files_in_multipart_content(
    sample_jsonl_file, sample_csv_file, sample_json_file, tmp_path
):
    files = (sample_jsonl_file, sample_csv_file, sample_json_file)

    tmpdir = str(tmp_path)
    multipart_file = os.path.join(tmpdir, "multipart")
    boundary_string = "foobar"

    with open(multipart_file, "w") as fp:
        for file in files:
            fp.write("--")
            fp.write(boundary_string)
            fp.writelines(
                [
                    "\n",
                    f'Content-Disposition: form-data; name="{file}"; filename="{file}"\n',
                    "Content-Type: text/plain\n",
                    "\n",
                ]
            )
            with open(file, "r") as inner_fp:
                content = inner_fp.read()
            fp.write(content)
            fp.write("\n")

        fp.write("--")
        fp.write(boundary_string)
        fp.write("--")

    parts = list(multipart_decoder.stream_multipart_file(multipart_file))

    assert len(parts) == len(files)

    for actual, expected in zip(parts, files):
        actual_content = actual.fp.read()
        with open(expected, "rb") as f:
            expected_content = f.read()
        assert actual_content == expected_content



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common data model objects used to identify the producer of a given output
"""

# Local
from .dataobject import DataObjectBase, dataobject
from .package import PACKAGE_COMMON


@dataobject(PACKAGE_COMMON)
class ProducerId(DataObjectBase):
    """Information about a data structure and the module that produced it."""

    name: str
    version: str

    def __add__(self, other):
        """Add two producer ids."""
        return ProducerId(name=" & ".join([self.name, other.name]), version="0.0.0")

    @classmethod
    def from_proto(cls, proto):
        """Overloaded implementation for efficiency vs base introspection"""
        return cls(proto.name, proto.version)

    def fill_proto(self, proto):
        """Overloaded implementation for efficiency vs base introspection"""
        if self.name is not None:
            proto.name = self.name
        if self.version is not None:
            proto.version = self.version
        return proto



# Local
from .file_processing import BoundingBoxModule
from .geospatial import GeoStreamingModule
from .other_task import OtherModule
from .sample_task import (
    CompositeModule,
    InnerModule,
    ListModule,
    SampleModule,
    SamplePrimitiveModule,
    StreamingModule,
)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Augmentor representing multiple augmentor behaviors combined into one configurable class.
"""
# First Party
import alog

# Local
from ..exceptions import error_handler
from .base import AugmentorBase
from .schemes import SchemeBase

log = alog.use_channel("MRGD_AUGMNTR")
error = error_handler.get(log)


class MergedAugmentor(AugmentorBase):
    def __init__(self, scheme):
        """Build an augmentor encapsulating multiple augmentors, where application order is
        governed by the provided scheme.
        Args:
            scheme (SchemeBase): Scheme indicating how encapsulated augmentors
                should be combined.
        """
        # NOTE: Random seed of merged augmentor does not currently matter since randomness is
        # already represented completely within encapsulated augmentors and scheme state
        super().__init__(random_seed=1001)
        error.type_check("<COR10421239E>", SchemeBase, scheme=scheme)
        self._scheme = scheme
        # pylint: disable=consider-using-set-comprehension
        augmentor_types = set([aug.augmentor_type for aug in self._scheme._augmentors])
        error.value_check(
            "<COR18849029E>",
            len(augmentor_types) == 1,
            "Cannot merge augmentors with differing augmentor types",
        )
        error.value_check(
            "<COR18146529E>",
            all(not aug.produces_none for aug in self._scheme._augmentors),
            "Cannot merge augmentors that produces <None> outputs",
        )
        self.augmentor_type = augmentor_types.pop()

    def _augment(self, obj):
        """Apply a merged augmentor whose behavior is controlled by the encapsulated scheme.

        Args:
            obj (str | caikit.core.data_model.DataBase): Object to be augmented.
        Returns:
            str | caikit.core.data_model.DataBase: Augmented object of same type
                as input obj.
        """
        return self._scheme.execute(obj)

    def reset(self):
        """Reset the merge augmentor's random number generation. In this case we actually don't
        need to care about the augmentor random state; what matters for this is the augmentor.
        scheme.
        """
        self._scheme.reset()
